#A 'Binney' quasi-isothermal DF
import math
import warnings
import numpy
import hashlib
from scipy import optimize, interpolate, integrate
from galpy import potential
from galpy import actionAngle
from galpy.actionAngle import actionAngleIsochrone
from galpy.potential import IsochronePotential
from galpy.potential import flatten as flatten_potential
from galpy.orbit import Orbit
from .df import df, _APY_LOADED
from galpy.util import galpyWarning
from galpy.util.bovy_conversion import physical_conversion, \
    potential_physical_input, actionAngle_physical_input, _APY_UNITS
if _APY_LOADED:
    from astropy import units
_NSIGMA=4
_DEFAULTNGL=10
_DEFAULTNGL2=20
class quasiisothermaldf(df):
    """Class that represents a 'Binney' quasi-isothermal DF"""
    def __init__(self,hr,sr,sz,hsr,hsz,pot=None,aA=None,
                 cutcounter=False,
                 _precomputerg=True,_precomputergrmax=None,
                 _precomputergnLz=51,
                 refr=1.,lo=10./220./8.,
                 ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           Initialize a quasi-isothermal DF

        INPUT:

           hr - radial scale length (can be Quantity)

           sr - radial velocity dispersion at the solar radius (can be Quantity)

           sz - vertical velocity dispersion at the solar radius (can be Quantity)

           hsr - radial-velocity-dispersion scale length (can be Quantity)

           hsz - vertial-velocity-dispersion scale length (can be Quantity)

           pot= Potential instance or list thereof

           aA= actionAngle instance used to convert (x,v) to actions [must be an instance of an actionAngle class that computes (J,Omega,angle) for a given (x,v)]

           cutcounter= if True, set counter-rotating stars' DF to zero

           refr= reference radius for dispersions (can be different from ro) (can be Quantity)

           lo= reference angular momentum below where there are significant numbers of retrograde stars (can be Quantity)

           ro= distance from vantage point to GC (kpc; can be Quantity)

           vo= circular velocity at ro (km/s; can be Quantity)

        OTHER INPUTS:

           _precomputerg= if True (default), pre-compute the rL(L)

           _precomputergrmax= if set, this is the maximum R for which to pre-compute rg (default: 5*hr)

           _precomputergnLz if set, number of Lz to pre-compute rg for (default: 51)

        OUTPUT:

           object

        HISTORY:

           2012-07-25 - Started - Bovy (IAS@MPIA)

        """
        df.__init__(self,ro=ro,vo=vo)
        if _APY_LOADED and isinstance(hr,units.Quantity):
            hr= hr.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(sr,units.Quantity):
            sr= sr.to(units.km/units.s).value/self._vo
        if _APY_LOADED and isinstance(sz,units.Quantity):
            sz= sz.to(units.km/units.s).value/self._vo
        if _APY_LOADED and isinstance(hsr,units.Quantity):
            hsr= hsr.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(hsz,units.Quantity):
            hsz= hsz.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(refr,units.Quantity):
            refr= refr.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(lo,units.Quantity):
            lo= lo.to(units.kpc*units.km/units.s).value/self._ro/self._vo
        self._hr= hr
        self._sr= sr
        self._sz= sz
        self._hsr= hsr
        self._hsz= hsz
        self._refr= refr
        self._lo= lo
        self._lnsr= math.log(self._sr)
        self._lnsz= math.log(self._sz)
        self._maxVT_hash= None
        self._maxVT_ip= None
        if pot is None:
            raise IOError("pot= must be set")
        self._pot= flatten_potential(pot)
        if aA is None:
            raise IOError("aA= must be set")
        self._aA= aA
        if not self._aA._pot == self._pot:
            if not isinstance(self._aA,actionAngleIsochrone):
                raise IOError("Potential in aA does not appear to be the same as given potential pot")
            elif isinstance(self._pot,IsochronePotential) and \
                    not self._aA.b == self._pot.b and \
                    not self._aA.amp == self._pot._amp:
                raise IOError("Potential in aA does not appear to be the same as given potential pot")
        self._cutcounter= cutcounter
        if _precomputerg:
            if _precomputergrmax is None:
                _precomputergrmax= 5*self._hr
            self._precomputergrmax= _precomputergrmax
            self._precomputergnLz= _precomputergnLz
            self._precomputergLzmin= 0.01
            self._precomputergLzmax= self._precomputergrmax\
                *potential.vcirc(self._pot,self._precomputergrmax)
            self._precomputergLzgrid= numpy.linspace(self._precomputergLzmin,self._precomputergLzmax,self._precomputergnLz)
            self._rls= numpy.array([potential.rl(self._pot,l) for l in self._precomputergLzgrid])
            #Spline interpolate
            self._rgInterp= interpolate.InterpolatedUnivariateSpline(self._precomputergLzgrid,self._rls,k=3)
        else:
            self._precomputergrmax= 0.
            self._rgInterp= None
            self._rls= None
            self._precomputergnr= None
            self._precomputergLzgrid= None
            self._precomputergLzmin= \
                numpy.finfo(numpy.dtype(numpy.float64)).max
            self._precomputergLzmax= \
                numpy.finfo(numpy.dtype(numpy.float64)).min
        self._precomputerg= _precomputerg
        self._glxdef, self._glwdef= \
            numpy.polynomial.legendre.leggauss(_DEFAULTNGL)
        self._glxdef2, self._glwdef2= \
            numpy.polynomial.legendre.leggauss(_DEFAULTNGL2)
        self._glxdef12, self._glwdef12= \
            numpy.polynomial.legendre.leggauss(_DEFAULTNGL//2)
        return None

    @physical_conversion('phasespacedensity',pop=True)
    def __call__(self,*args,**kwargs):
        """
        NAME:

           __call__

        PURPOSE:

           return the DF

        INPUT:

           Either:

              a)(jr,lz,jz) tuple; each can be a Quantity
                 where:
                    jr - radial action
                    lz - z-component of angular momentum
                    jz - vertical action

              b) R,vR,vT,z,vz

              c) Orbit instance: initial condition used if that's it, orbit(t)
                 if there is a time given as well

           log= if True, return the natural log

           +scipy.integrate.quadrature kwargs

           func= function of (jr,lz,jz) to multiply f with (useful for moments)

        OUTPUT:

           value of DF

        HISTORY:

           2012-07-25 - Written - Bovy (IAS@MPIA)

        NOTE:

           For Miyamoto-Nagai/adiabatic approximation this seems to take 
           about 30 ms / evaluation in the extended Solar neighborhood
           For a MWPotential/adiabatic approximation this takes about 
           50 ms / evaluation in the extended Solar neighborhood

           For adiabatic-approximation grid this seems to take 
           about 0.67 to 0.75 ms / evaluation in the extended Solar 
           neighborhood (includes some out of the grid)

           up to 200x faster when called with vector R,vR,vT,z,vz

        """
        #First parse log
        log= kwargs.pop('log',False)
        _return_actions= kwargs.pop('_return_actions',False)
        _return_freqs= kwargs.pop('_return_freqs',False)
        _func= kwargs.pop('func',None)
        if 'rg' in kwargs:
            thisrg= kwargs.pop('rg')
            kappa= kwargs.pop('kappa')
            nu= kwargs.pop('nu')
            Omega= kwargs.pop('Omega')
        else:
            thisrg= None
            kappa= None
            nu= None
            Omega= None
        #First parse args
        if len(args) == 1 and not isinstance(args[0],Orbit): #(jr,lz,jz)
            jr,lz,jz= args[0]
            if _APY_LOADED and isinstance(jr,units.Quantity):
                jr= jr.to(units.kpc*units.km/units.s).value/self._ro/self._vo
            if _APY_LOADED and isinstance(lz,units.Quantity):
                lz= lz.to(units.kpc*units.km/units.s).value/self._ro/self._vo
            if _APY_LOADED and isinstance(jz,units.Quantity):
                jz= jz.to(units.kpc*units.km/units.s).value/self._ro/self._vo
        else:
            #Use self._aA to calculate the actions
            if isinstance(args[0],Orbit) and len(args[0].shape) > 1:
                raise RuntimeError("Evaluating quasiisothermaldf with Orbit instances with multi-dimensional shapes is not supported") #pragma: no cover
            try:
                jr,lz,jz= self._aA(*args,use_physical=False,**kwargs)
            except actionAngle.UnboundError:
                if log: return -numpy.finfo(numpy.dtype(numpy.float64)).max
                else: return 0.
            #if isinstance(jr,(list,numpy.ndarray)) and len(jr) > 1: jr= jr[0]
            #if isinstance(jz,(list,numpy.ndarray)) and len(jz) > 1: jz= jz[0]
        if not isinstance(lz,numpy.ndarray) and self._cutcounter and lz < 0.:
            if log: return -numpy.finfo(numpy.dtype(numpy.float64)).max
            else: return 0.
        #First calculate rg
        if thisrg is None:
            thisrg= self._rg(lz)
            #Then calculate the epicycle and vertical frequencies
            kappa, nu= self._calc_epifreq(thisrg), self._calc_verticalfreq(thisrg)
            Omega= numpy.fabs(lz)/thisrg/thisrg
        #calculate surface-densities and sigmas
        lnsurfmass= (self._refr-thisrg)/self._hr
        lnsr= self._lnsr+(self._refr-thisrg)/self._hsr
        lnsz= self._lnsz+(self._refr-thisrg)/self._hsz
        #Calculate func
        if not _func is None:
            if log:
                funcTerm= numpy.log(_func(jr,lz,jz))
            else:
                funcFactor= _func(jr,lz,jz)
        #Calculate fsr
        else:
            if log:
                funcTerm= 0.
            else:
                funcFactor= 1.            
        if log:
            lnfsr= numpy.log(Omega)+lnsurfmass-2.*lnsr-math.log(math.pi)\
                -numpy.log(kappa)\
                +numpy.log(1.+numpy.tanh(lz/self._lo))\
                -kappa*jr*numpy.exp(-2.*lnsr)
            lnfsz= numpy.log(nu)-math.log(2.*math.pi)\
                -2.*lnsz-nu*jz*numpy.exp(-2.*lnsz)
            out= lnfsr+lnfsz+funcTerm
            if isinstance(lz,numpy.ndarray):
                out[numpy.isnan(out)]= -numpy.finfo(numpy.dtype(numpy.float64)).max
                if self._cutcounter: out[(lz < 0.)]= -numpy.finfo(numpy.dtype(numpy.float64)).max
            elif numpy.isnan(out): out= -numpy.finfo(numpy.dtype(numpy.float64)).max
        else:
            srm2= numpy.exp(-2.*lnsr)
            fsr= Omega*numpy.exp(lnsurfmass)*srm2/math.pi/kappa\
                *(1.+numpy.tanh(lz/self._lo))\
                *numpy.exp(-kappa*jr*srm2)
            szm2= numpy.exp(-2.*lnsz)
            fsz= nu/2./math.pi*szm2*numpy.exp(-nu*jz*szm2)
            out= fsr*fsz*funcFactor
            if isinstance(lz,numpy.ndarray):
                out[numpy.isnan(out)]= 0.
                if self._cutcounter: out[(lz < 0.)]= 0.
            elif numpy.isnan(out): out= 0.
        if _return_actions and _return_freqs:
            return (out,jr,lz,jz,thisrg,kappa,nu,Omega)
        elif _return_actions:
            return (out,jr,lz,jz)
        elif _return_freqs:
            return (out,thisrg,kappa,nu,Omega)
        else:
            return out

    @potential_physical_input
    @physical_conversion('position',pop=True)
    def estimate_hr(self,R,z=0.,dR=10.**-8.,**kwargs):
        """
        NAME:

           estimate_hr

        PURPOSE:

           estimate the exponential scale length at R

        INPUT:

           R - Galactocentric radius (can be Quantity)

           z= height (default: 0 pc) (can be Quantity)

           dR- range in R to use (can be Quantity)

           density kwargs

        OUTPUT:

           estimated hR

        HISTORY:

           2012-09-11 - Written - Bovy (IAS)

           2013-01-28 - Re-written - Bovy

        """
        Rs= [R-dR/2.,R+dR/2.]
        if z is None:
            sf= numpy.array([self.surfacemass_z(r,use_physical=False,
                                                **kwargs) for r in Rs])
        else:
            sf= numpy.array([self.density(r,z,use_physical=False,
                                          **kwargs) for r in Rs])
        lsf= numpy.log(sf)
        return -dR/(lsf[1]-lsf[0])

    @potential_physical_input
    @physical_conversion('position',pop=True)
    def estimate_hz(self,R,z,dz=10.**-8.,**kwargs):
        """
        NAME:

           estimate_hz

        PURPOSE:

           estimate the exponential scale height at R

        INPUT:

           R - Galactocentric radius (can be Quantity)

           dz - z range to use (can be Quantity)

           density kwargs

        OUTPUT:

           estimated hz

        HISTORY:

           2012-08-30 - Written - Bovy (IAS)

           2013-01-28 - Re-written - Bovy

        """
        if z == 0.:
            zs= [z,z+dz]
        else:
            zs= [z-dz/2.,z+dz/2.]
        sf= numpy.array([self.density(R,zz,use_physical=False,
                                      **kwargs) for zz in zs])
        lsf= numpy.log(sf)
        return -dz/(lsf[1]-lsf[0])

    @potential_physical_input
    @physical_conversion('position',pop=True)
    def estimate_hsr(self,R,z=0.,dR=10.**-8.,**kwargs):
        """
        NAME:

           estimate_hsr

        PURPOSE:

           estimate the exponential scale length of the radial dispersion at R

        INPUT:

           R - Galactocentric radius (can be Quantity)

           z= height (default: 0 pc) (can be Quantity)

           dR- range in R to use (can be Quantity)

           density kwargs

        OUTPUT:

           estimated hsR

        HISTORY:

           2013-03-08 - Written - Bovy (IAS)

        """
        Rs= [R-dR/2.,R+dR/2.]
        sf= numpy.array([self.sigmaR2(r,z,use_physical=False,
                                      **kwargs) for r in Rs])
        lsf= numpy.log(sf)/2.
        return -dR/(lsf[1]-lsf[0])

    @potential_physical_input
    @physical_conversion('position',pop=True)
    def estimate_hsz(self,R,z=0.,dR=10.**-8.,**kwargs):
        """
        NAME:

           estimate_hsz

        PURPOSE:

           estimate the exponential scale length of the vertical dispersion at R

        INPUT:

           R - Galactocentric radius (can be Quantity)

           z= height (default: 0 pc) (can be Quantity)

           dR- range in R to use (can be Quantity)

           density kwargs

        OUTPUT:

           estimated hsz

        HISTORY:

           2013-03-08 - Written - Bovy (IAS)

        """
        Rs= [R-dR/2.,R+dR/2.]
        sf= numpy.array([self.sigmaz2(r,z,use_physical=False,
                                      **kwargs) for r in Rs])
        lsf= numpy.log(sf)/2.
        return -dR/(lsf[1]-lsf[0])

    @potential_physical_input
    @physical_conversion('numbersurfacedensity',pop=True)
    def surfacemass_z(self,R,nz=7,zmax=1.,fixed_quad=True,fixed_order=8,
                      **kwargs):
        """
        NAME:

           surfacemass_z

        PURPOSE:

           calculate the vertically-integrated surface density

        INPUT:

           R - Galactocentric radius (can be Quantity)

           fixed_quad= if True (default), use Gauss-Legendre integration

           fixed_order= (20), order of GL integration to use

           nz= number of zs to use to estimate

           zmax= maximum z to use (can be Quantity)

           density kwargs

        OUTPUT:

           \Sigma(R)

        HISTORY:

           2012-08-30 - Written - Bovy (IAS)

        """
        if fixed_quad:
            return 2.*integrate.fixed_quad(lambda x: self.density(R*numpy.ones(fixed_order),x,use_physical=False),
                                           0.,.5,n=fixed_order)[0]
        zs= numpy.linspace(0.,zmax,nz)
        sf= numpy.array([self.density(R,z,use_physical=False,
                                      **kwargs) for z in zs])
        lsf= numpy.log(sf)
        #Interpolate
        lsfInterp= interpolate.UnivariateSpline(zs,
                                                lsf,
                                                k=3)
        #Integrate
        return 2.*integrate.quad((lambda x: numpy.exp(lsfInterp(x))),
                                 0.,1.)[0]

    def vmomentdensity(self,*args,**kwargs):
        """
        NAME:

           vmomentdensity

        PURPOSE:

           calculate the an arbitrary moment of the velocity distribution 
           at R times the density

        INPUT:

           R - radius at which to calculate the moment(/ro)

           n - vR^n

           m - vT^m

           o - vz^o

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the vR and vz velocities over (when doing explicit numerical integral; default: 4)
           
           vTmax - upper limit for integration over vT (default: 1.5)

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= use Gauss-Legendre

           _returngl= if True, return the evaluated DF

           _return_actions= if True, return the evaluated actions (does not work with _returngl currently)

           _return_freqs= if True, return the evaluated frequencies and rg (does not work with _returngl currently)

        OUTPUT:

           <vR^n vT^m  x density> at R,z (no support for units)

        HISTORY:

           2012-08-06 - Written - Bovy (IAS@MPIA)

        """
        use_physical= kwargs.pop('use_physical',True)
        ro= kwargs.pop('ro',None)
        if ro is None and hasattr(self,'_roSet') and self._roSet:
            ro= self._ro
        if _APY_LOADED and isinstance(ro,units.Quantity):
            ro= ro.to(units.kpc).value
        vo= kwargs.pop('vo',None)
        if vo is None and hasattr(self,'_voSet') and self._voSet:
            vo= self._vo
        if _APY_LOADED and isinstance(vo,units.Quantity):
            vo= vo.to(units.km/units.s).value
        if use_physical and not vo is None and not ro is None:
            fac= vo**(args[2]+args[3]+args[4])/ro**3
            if _APY_UNITS:
                u= 1/units.kpc**3*(units.km/units.s)**(args[2]+args[3]+args[4])
            out= self._vmomentdensity(*args,**kwargs)
            if _APY_UNITS:
                return units.Quantity(out*fac,unit=u)
            else:
                return out*fac
        else:
            return self._vmomentdensity(*args,**kwargs)  

    def _vmomentdensity(self,R,z,n,m,o,nsigma=None,mc=False,nmc=10000,
                       _returnmc=False,_vrs=None,_vts=None,_vzs=None,
                       _rawgausssamples=False,
                       gl=False,ngl=_DEFAULTNGL,_returngl=False,_glqeval=None,
                       _return_actions=False,_jr=None,_lz=None,_jz=None,
                       _return_freqs=False,
                       _rg=None,_kappa=None,_nu=None,_Omega=None,
                       _sigmaR1=None,_sigmaz1=None,
                       **kwargs):
        """Non-physical version of vmomentdensity, otherwise the same"""
        if isinstance(R,numpy.ndarray):
            return numpy.array([self._vmomentdensity(r,zz,n,m,o,nsigma=nsigma,
                                                    mc=mc,nmc=nmc,
                                                    gl=gl,ngl=ngl,**kwargs) for r,zz in zip(R,z)])
        if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                actionAngle.actionAngleAdiabaticGrid)):
            if n % 2 == 1. or o % 2 == 1.:
                return 0. #we know this must be the case
        if nsigma == None:
            nsigma= _NSIGMA
        if _sigmaR1 is None:
            sigmaR1= self._sr*numpy.exp((self._refr-R)/self._hsr)
        else:
            sigmaR1= _sigmaR1
        if _sigmaz1 is None:
            sigmaz1= self._sz*numpy.exp((self._refr-R)/self._hsz)
        else:
            sigmaz1= _sigmaz1
        thisvc= potential.vcirc(self._pot,R,use_physical=False)
        #Use the asymmetric drift equation to estimate va
        gamma= numpy.sqrt(0.5)
        va= sigmaR1**2./2./thisvc\
            *(gamma**2.-1. #Assume close to flat rotation curve, sigphi2/sigR2 =~ 0.5
               +R*(1./self._hr+2./self._hsr))
        if math.fabs(va) > sigmaR1: va = 0.#To avoid craziness near the center
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            if not _glqeval is None and ngl != _glqeval.shape[0]:
                _glqeval= None
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vRgl= nsigma*sigmaR1/2.*(glx+1.)
                vzgl= nsigma*sigmaz1/2.*(glx+1.)
                vRglw= glw
                vzglw= glw
            else:
                vRgl= nsigma*sigmaR1/2.*(glx12+1.)
                #vRgl= 1.5/2.*(glx12+1.)
                vRgl= list(vRgl)
                vRgl.extend(-nsigma*sigmaR1/2.*(glx12+1.))
                #vRgl.extend(-1.5/2.*(glx12+1.))
                vRgl= numpy.array(vRgl)
                vzgl= nsigma*sigmaz1/2.*(glx12+1.)
                #vzgl= 1.5/2.*(glx12+1.)
                vzgl= list(vzgl)
                vzgl.extend(-nsigma*sigmaz1/2.*(glx12+1.))
                #vzgl.extend(-1.5/2.*(glx12+1.))
                vzgl= numpy.array(vzgl)
                vRglw= glw12
                vRglw= list(vRglw)
                vRglw.extend(glw12)
                vRglw= numpy.array(vRglw)
                vzglw= glw12
                vzglw= list(vzglw)
                vzglw.extend(glw12)
                vzglw= numpy.array(vzglw)
            if 'vTmax' in kwargs: vTmax = kwargs['vTmax']
            else:                 vTmax = 1.5
            vTgl= vTmax/2.*(glx+1.)
            #Tile everything
            vTgl= numpy.tile(vTgl,(ngl,ngl,1)).T
            vRgl= numpy.tile(numpy.reshape(vRgl,(1,ngl)).T,(ngl,1,ngl))
            vzgl= numpy.tile(vzgl,(ngl,ngl,1))
            vTglw= numpy.tile(glw,(ngl,ngl,1)).T #also tile weights
            vRglw= numpy.tile(numpy.reshape(vRglw,(1,ngl)).T,(ngl,1,ngl))
            vzglw= numpy.tile(vzglw,(ngl,ngl,1))
            #evaluate
            if _glqeval is None and _jr is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self(R+numpy.zeros(ngl*ngl*ngl),
                                           vRgl.flatten(),
                                           vTgl.flatten(),
                                           z+numpy.zeros(ngl*ngl*ngl),
                                           vzgl.flatten(),
                                           log=True,
                                           _return_actions=True,
                                           _return_freqs=True,
                                           use_physical=False)
                logqeval= numpy.reshape(logqeval,(ngl,ngl,ngl))
            elif not _jr is None and _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self((_jr,_lz,_jz),
                                                                 log=True,
                                                                 _return_actions=True,
                                                                 _return_freqs=True,
                                                                 use_physical=False)
                logqeval= numpy.reshape(logqeval,(ngl,ngl,ngl))
            elif not _jr is None and not _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self((_jr,_lz,_jz),
                                           rg=_rg,kappa=_kappa,nu=_nu,
                                           Omega=_Omega,
                                           log=True,
                                           _return_actions=True,
                                           _return_freqs=True,
                                           use_physical=False)
                logqeval= numpy.reshape(logqeval,(ngl,ngl,ngl))
            else:
                logqeval= _glqeval
            if _returngl:
                return (numpy.sum(numpy.exp(logqeval)*vRgl**n*vTgl**m*vzgl**o
                                  *vTglw*vRglw*vzglw)*sigmaR1*sigmaz1*0.125*vTmax*nsigma**2,
                        logqeval)
            elif _return_actions and _return_freqs:
                return (numpy.sum(numpy.exp(logqeval)*vRgl**n*vTgl**m*vzgl**o
                                  *vTglw*vRglw*vzglw)*sigmaR1*sigmaz1*0.125*vTmax*nsigma**2,
                        jr,lz,jz,
                        rg,kappa,nu,Omega)
            elif _return_actions:
                return (numpy.sum(numpy.exp(logqeval)*vRgl**n*vTgl**m*vzgl**o
                                  *vTglw*vRglw*vzglw)*sigmaR1*sigmaz1*0.125*vTmax*nsigma**2,
                        jr,lz,jz)
            else:
                return numpy.sum(numpy.exp(logqeval)*vRgl**n*vTgl**m*vzgl**o
                                 *vTglw*vRglw*vzglw*sigmaR1*sigmaz1*0.125*vTmax*nsigma**2)
        elif mc:
            mvT= (thisvc-va)/gamma/sigmaR1
            if _vrs is None:
                vrs= numpy.random.normal(size=nmc)
            else:
                vrs= _vrs
            if _vts is None:
                vts= numpy.random.normal(size=nmc)+mvT
            else:
                if _rawgausssamples:
                    vts= _vts+mvT
                else:
                    vts= _vts
            if _vzs is None:
                vzs= numpy.random.normal(size=nmc)
            else:
                vzs= _vzs
            Is= _vmomentsurfaceMCIntegrand(vzs,vrs,vts,numpy.ones(nmc)*R,
                                           numpy.ones(nmc)*z,
                                           self,sigmaR1,gamma,sigmaz1,mvT,
                                           n,m,o)
            if _returnmc:
                if _rawgausssamples:
                    return (numpy.mean(Is)*sigmaR1**(2.+n+m)*gamma**(1.+m)*sigmaz1**(1.+o),
                        vrs,vts-mvT,vzs)
                else:
                    return (numpy.mean(Is)*sigmaR1**(2.+n+m)*gamma**(1.+m)*sigmaz1**(1.+o),
                            vrs,vts,vzs)
            else:
                return numpy.mean(Is)*sigmaR1**(2.+n+m)*gamma**(1.+m)*sigmaz1**(1.+o)
        else: #pragma: no cover because this is too slow; a warning is shown
            warnings.warn("Calculations using direct numerical integration using tplquad is not recommended and extremely slow; it has also not been carefully tested",galpyWarning)
            return integrate.tplquad(_vmomentsurfaceIntegrand,
                                     1./gamma*(thisvc-va)/sigmaR1-nsigma,
                                     1./gamma*(thisvc-va)/sigmaR1+nsigma,
                                     lambda x: 0., lambda x: nsigma,
                                     lambda x,y: 0., lambda x,y: nsigma,
                                     (R,z,self,sigmaR1,gamma,sigmaz1,n,m,o),
                                     **kwargs)[0]*sigmaR1**(2.+n+m)*gamma**(1.+m)*sigmaz1**(1.+o)
        
    def jmomentdensity(self,*args,**kwargs):
        """
        NAME:

           jmomentdensity
        PURPOSE:

           calculate the an arbitrary moment of an action
           of the velocity distribution 
           at R times the surfacmass
        INPUT:

           R - radius at which to calculate the moment(/ro)

           n - jr^n

           m - lz^m

           o - jz^o

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over (when doing explicit numerical integral)

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

        OUTPUT:

           <jr^n lz^m jz^o  x density> at R (no support for units)

        HISTORY:

           2012-08-09 - Written - Bovy (IAS@MPIA)

        """
        use_physical= kwargs.pop('use_physical',True)
        ro= kwargs.pop('ro',None)
        if ro is None and hasattr(self,'_roSet') and self._roSet:
            ro= self._ro
        if _APY_LOADED and isinstance(ro,units.Quantity):
            ro= ro.to(units.kpc).value
        vo= kwargs.pop('vo',None)
        if vo is None and hasattr(self,'_voSet') and self._voSet:
            vo= self._vo
        if _APY_LOADED and isinstance(vo,units.Quantity):
            vo= vo.to(units.km/units.s).value
        if use_physical and not vo is None and not ro is None:
            fac= (ro*vo)**(args[2]+args[3]+args[4])/ro**3
            if _APY_UNITS:
                u= 1/units.kpc**3*(units.kpc*units.km/units.s)**(args[2]+args[3]+args[4])
            out= self._jmomentdensity(*args,**kwargs)
            if _APY_UNITS:
                return units.Quantity(out*fac,unit=u)
            else:
                return out*fac
        else:
            return self._jmomentdensity(*args,**kwargs)  

    def _jmomentdensity(self,R,z,n,m,o,nsigma=None,mc=True,nmc=10000,
                       _returnmc=False,_vrs=None,_vts=None,_vzs=None,
                       **kwargs):
        """Non-physical version of jmomentdensity, otherwise the same"""
        if nsigma == None:
            nsigma= _NSIGMA
        sigmaR1= self._sr*numpy.exp((self._refr-R)/self._hsr)
        sigmaz1= self._sz*numpy.exp((self._refr-R)/self._hsz)
        thisvc= potential.vcirc(self._pot,R,use_physical=False)
        #Use the asymmetric drift equation to estimate va
        gamma= numpy.sqrt(0.5)
        va= sigmaR1**2./2./thisvc\
            *(gamma**2.-1. #Assume close to flat rotation curve, sigphi2/sigR2 =~ 0.5
               +R*(1./self._hr+2./self._hsr))
        if math.fabs(va) > sigmaR1: va = 0.#To avoid craziness near the center
        if mc:
            mvT= (thisvc-va)/gamma/sigmaR1
            if _vrs is None:
                vrs= numpy.random.normal(size=nmc)
            else:
                vrs= _vrs
            if _vts is None:
                vts= numpy.random.normal(size=nmc)+mvT
            else:
                vts= _vts
            if _vzs is None:
                vzs= numpy.random.normal(size=nmc)
            else:
                vzs= _vzs
            Is= _jmomentsurfaceMCIntegrand(vzs,vrs,vts,numpy.ones(nmc)*R,numpy.ones(nmc)*z,self,sigmaR1,gamma,sigmaz1,mvT,n,m,o)
            if _returnmc:
                return (numpy.mean(Is)*sigmaR1**2.*gamma*sigmaz1,
                        vrs,vts,vzs)
            else:
                return numpy.mean(Is)*sigmaR1**2.*gamma*sigmaz1
        else: #pragma: no cover because this is too slow; a warning is shown
            warnings.warn("Calculations using direct numerical integration using tplquad is not recommended and extremely slow; it has also not been carefully tested",galpyWarning)
            return integrate.tplquad(_jmomentsurfaceIntegrand,
                                     1./gamma*(thisvc-va)/sigmaR1-nsigma,
                                     1./gamma*(thisvc-va)/sigmaR1+nsigma,
                                     lambda x: 0., lambda x: nsigma,
                                     lambda x,y: 0., lambda x,y: nsigma,
                                     (R,z,self,sigmaR1,gamma,sigmaz1,n,m,o),
                                     **kwargs)[0]*sigmaR1**2.*gamma*sigmaz1
        
    @potential_physical_input
    @physical_conversion('numberdensity',pop=True)
    def density(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           density

        PURPOSE:

           calculate the density at R,z by marginalizing over velocity

        INPUT:

           R - radius at which to calculate the density (can be Quantity)

           z - height at which to calculate the density (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           density at (R,z)

        HISTORY:

           2012-07-26 - Written - Bovy (IAS@MPIA)

        """
        return self._vmomentdensity(R,z,0.,0.,0.,
                                   nsigma=nsigma,mc=mc,nmc=nmc,
                                   gl=gl,ngl=ngl,
                                   **kwargs)
    
    @potential_physical_input
    @physical_conversion('velocity2',pop=True)
    def sigmaR2(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           sigmaR2

        PURPOSE:

           calculate sigma_R^2 by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           sigma_R^2

        HISTORY:

           2012-07-30 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,2.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,2.,0.,0.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,2.,0.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('velocity2',pop=True)
    def sigmaRz(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           sigmaRz

        PURPOSE:

           calculate sigma_RZ^2 by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           sigma_Rz^2

        HISTORY:

           2012-07-30 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,1.,0.,1.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,1.,0.,1.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,1.,0.,1.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('angle',pop=True)
    def tilt(self,R,z,nsigma=None,mc=False,nmc=10000,
             gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           tilt

        PURPOSE:

           calculate the tilt of the velocity ellipsoid by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           tilt in rad

        HISTORY:

           2012-12-23 - Written - Bovy (IAS)

           2017-10-28 - Changed return unit to rad - Bovy (UofT)

        """
        warnings.warn("In versions >1.3, the output unit of quasiisothermaldf.tilt has been changed to radian (from degree before)",galpyWarning)
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            tsigmar2= self._vmomentdensity(R,z,2.,0.,0.,
                                              nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                              _vrs=vrs,_vts=vts,_vzs=vzs,
                                              **kwargs)/surfmass
            tsigmaz2= self._vmomentdensity(R,z,0.,0.,2.,
                                              nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                              _vrs=vrs,_vts=vts,_vzs=vzs,
                                              **kwargs)/surfmass
            tsigmarz= self._vmomentdensity(R,z,1.,0.,1.,
                                              nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                              _vrs=vrs,_vts=vts,_vzs=vzs,
                                              **kwargs)/surfmass
            return 0.5*numpy.arctan(2.*tsigmarz/(tsigmar2-tsigmaz2))
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            tsigmar2= self._vmomentdensity(R,z,2.,0.,0.,
                                             ngl=ngl,gl=gl,
                                             _glqeval=glqeval,
                                             **kwargs)/surfmass
            tsigmaz2= self._vmomentdensity(R,z,0.,0.,2.,
                                              ngl=ngl,gl=gl,
                                              _glqeval=glqeval,
                                              **kwargs)/surfmass
            tsigmarz= self._vmomentdensity(R,z,1.,0.,1.,
                                              ngl=ngl,gl=gl,
                                              _glqeval=glqeval,
                                              **kwargs)/surfmass
            return 0.5*numpy.arctan(2.*tsigmarz/(tsigmar2-tsigmaz2))
        else:
            raise NotImplementedError("Use either mc=True or gl=True")
        
    @potential_physical_input
    @physical_conversion('velocity2',pop=True)
    def sigmaz2(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           sigmaz2

        PURPOSE:

           calculate sigma_z^2 by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           sigma_z^2

        HISTORY:

           2012-07-30 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,0.,0.,2.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,0.,0.,2.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,0.,0.,2.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('velocity',pop=True)
    def meanvT(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           meanvT

        PURPOSE:

           calculate the mean rotational velocity by marginalizing over velocity 

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           meanvT

        HISTORY:

           2012-07-30 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,0.,1.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,0.,1.,0.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,0.,1.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('velocity',pop=True)
    def meanvR(self,R,z,nsigma=None,mc=False,nmc=10000,
               gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           meanvR

        PURPOSE:

           calculate the mean radial velocity by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           meanvR

        HISTORY:

           2012-12-23 - Written - Bovy (IAS)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,1.,0.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,1.,0.,0.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,1.,0.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('velocity',pop=True)
    def meanvz(self,R,z,nsigma=None,mc=False,nmc=10000,
               gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           meanvz

        PURPOSE:

           calculate the mean vertical velocity by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           meanvz

        HISTORY:

           2012-12-23 - Written - Bovy (IAS)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._vmomentdensity(R,z,0.,0.,1.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            return self._vmomentdensity(R,z,0.,0.,1.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._vmomentdensity(R,z,0.,0.,1.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('velocity2',pop=True)
    def sigmaT2(self,R,z,nsigma=None,mc=False,nmc=10000,
                gl=True,ngl=_DEFAULTNGL,**kwargs):
        """
        NAME:

           sigmaT2

        PURPOSE:

           calculate sigma_T^2 by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

           gl= if True, calculate using Gauss-Legendre integration

           ngl= if gl, use ngl-th order Gauss-Legendre integration for each dimension

        OUTPUT:

           sigma_T^2

        HISTORY:

           2012-07-30 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            mvt= self._vmomentdensity(R,z,0.,1.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
            return self._vmomentdensity(R,z,0.,2.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                           **kwargs)/surfmass\
                                           -mvt**2.
        elif gl:
            surfmass, glqeval= self._vmomentdensity(R,z,0.,0.,0.,
                                                       gl=gl,ngl=ngl,
                                                       _returngl=True,
                                                       **kwargs)
            mvt= self._vmomentdensity(R,z,0.,1.,0.,
                                         ngl=ngl,gl=gl,
                                         _glqeval=glqeval,
                                         **kwargs)/surfmass
            return self._vmomentdensity(R,z,0.,2.,0.,
                                           ngl=ngl,gl=gl,
                                           _glqeval=glqeval,
                                           **kwargs)/surfmass-mvt**2.

        else: #pragma: no cover because this is too slow; a warning is shown
            surfmass= self._vmomentdensity(R,z,0.,0.,0.,
                                              nsigma=nsigma,mc=mc,nmc=nmc,
                                              **kwargs)
            return (self._vmomentdensity(R,z,0.,2.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/surfmass\
                        -(self._vmomentdensity(R,z,0.,2.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/surfmass)**2.)

    @potential_physical_input
    @physical_conversion('action',pop=True)
    def meanjr(self,R,z,nsigma=None,mc=True,nmc=10000,**kwargs):
        """
        NAME:

           meanjr

        PURPOSE:

           calculate the mean radial action by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

        OUTPUT:

           meanjr

        HISTORY:

           2012-08-09 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._jmomentdensity(R,z,1.,0.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._jmomentdensity(R,z,1.,0.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('action',pop=True)
    def meanlz(self,R,z,nsigma=None,mc=True,nmc=10000,**kwargs):
        """
        NAME:

           meanlz

        PURPOSE:

           calculate the mean angular momemtum by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

        OUTPUT:

           meanlz

        HISTORY:

           2012-08-09 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._jmomentdensity(R,z,0.,1.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._jmomentdensity(R,z,0.,1.,0.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    @physical_conversion('action',pop=True)
    def meanjz(self,R,z,nsigma=None,mc=True,nmc=10000,**kwargs):
        """
        NAME:

           meanjz

        PURPOSE:

           calculate the mean vertical action by marginalizing over velocity

        INPUT:

           R - radius at which to calculate this (can be Quantity)

           z - height at which to calculate this (can be Quantity)

        OPTIONAL INPUT:

           nsigma - number of sigma to integrate the velocities over

           scipy.integrate.tplquad kwargs epsabs and epsrel

           mc= if True, calculate using Monte Carlo integration

           nmc= if mc, use nmc samples

        OUTPUT:

           meanjz

        HISTORY:

           2012-08-09 - Written - Bovy (IAS@MPIA)

        """
        if mc:
            surfmass, vrs, vts, vzs= self._vmomentdensity(R,z,0.,0.,0.,
                                                             nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=True,
                                                             **kwargs)
            return self._jmomentdensity(R,z,0.,0.,1.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,_returnmc=False,
                                           _vrs=vrs,_vts=vts,_vzs=vzs,
                                                             **kwargs)/surfmass
        else: #pragma: no cover because this is too slow; a warning is shown
            return (self._jmomentdensity(R,z,0.,0.,1.,
                                           nsigma=nsigma,mc=mc,nmc=nmc,
                                           **kwargs)/
                    self._vmomentdensity(R,z,0.,0.,0.,
                                            nsigma=nsigma,mc=mc,nmc=nmc,
                                            **kwargs))
        
    @potential_physical_input
    def sampleV(self,R,z,n=1,**kwargs):
        """
        NAME:

           sampleV

        PURPOSE:

           sample a radial, azimuthal, and vertical velocity at R,z

        INPUT:

           R - Galactocentric distance (can be Quantity)

           z - height (can be Quantity)

           n= number of distances to sample

        OUTPUT:

           list of samples

        HISTORY:

           2012-12-17 - Written - Bovy (IAS)

        """
        use_physical= kwargs.pop('use_physical',True)
        vo= kwargs.pop('vo',None)
        if vo is None and hasattr(self,'_voSet') and self._voSet:
            vo= self._vo
        if _APY_LOADED and isinstance(vo,units.Quantity):
            vo= vo.to(units.km/units.s).value
        #Determine the maximum of the velocity distribution
        maxVR= 0.
        maxVz= 0.
        maxVT= optimize.fmin_powell((lambda x: -self(R,0.,x,z,0.,log=True,
                                                     use_physical=False)),
                                    1.)
        logmaxVD= self(R,maxVR,maxVT,z,maxVz,log=True,use_physical=False)
        #Now rejection-sample
        vRs= []
        vTs= []
        vzs= []
        while len(vRs) < n:
            nmore= n-len(vRs)+1
            #sample
            propvR= numpy.random.normal(size=nmore)*2.*self._sr
            propvT= numpy.random.normal(size=nmore)*2.*self._sr+maxVT
            propvz= numpy.random.normal(size=nmore)*2.*self._sz
            VDatprop= self(R+numpy.zeros(nmore),
                           propvR,propvT,z+numpy.zeros(nmore),
                           propvz,log=True,use_physical=False)-logmaxVD
            VDatprop-= -0.5*(propvR**2./4./self._sr**2.+propvz**2./4./self._sz**2.\
                                 +(propvT-maxVT)**2./4./self._sr**2.)
            VDatprop= numpy.reshape(VDatprop,(nmore))
            indx= (VDatprop > numpy.log(numpy.random.random(size=nmore))) #accept
            vRs.extend(list(propvR[indx]))
            vTs.extend(list(propvT[indx]))
            vzs.extend(list(propvz[indx]))
        out= numpy.empty((n,3))
        out[:,0]= vRs[0:n]
        out[:,1]= vTs[0:n]
        out[:,2]= vzs[0:n]
        if use_physical and not vo is None:
            if _APY_UNITS:
                return units.Quantity(out*vo,unit=units.km/units.s)
            else:
                return out*vo
        else:
            return out

    @potential_physical_input
    def sampleV_interpolate(self,R,z,R_pixel,z_pixel,num_std=3,R_min=None,
                            R_max=None,z_max=None,**kwargs):
        """
        NAME:
            
            sampleV_interpolate
    
        PURPOSE:
            
            Given an array of R and z coordinates of stars, return the 
            positions and their radial, azimuthal, and vertical velocity.
    
        INPUT:
            
            R - array of Galactocentric distance (can be Quantity)

            z - array of height (can be Quantity)
            
            R_pixel, z_pixel= the pixel size for creating the grid for
                   interpolation (in natural unit)
            
            num_std= number of standard deviation to be considered outliers
                      sampled separately from interpolation
                      
            R_min, R_max, z_max= optional edges of the grid
    
        OUTPUT:
            
            coord_v= a numpy array containing the sampled velocity, (vR, vT, vz),
                     where each row correspond to the row of (R,z)
    
        HISTORY:
            
            2018-08-10 - Written - Samuel Wong (University of Toronto)
            
        """
        use_physical= kwargs.pop('use_physical',True)
        vo= kwargs.pop('vo',None)
        if vo is None and hasattr(self,'_voSet') and self._voSet:
            vo= self._vo
        if _APY_LOADED and isinstance(vo,units.Quantity):
            vo= vo.to(units.km/units.s).value
        #Initialize output array
        coord_v= numpy.empty((numpy.size(R), 3))
        #Since the sign of z doesn't matter, work with absolute value of z
        z= numpy.abs(z)
        # Grid edges
        if R_min is None:
            R_min= numpy.amax([numpy.mean(R)-num_std*numpy.std(R),
                               numpy.amin(R)])
        if R_max is None:
            R_max= numpy.amin([numpy.mean(R)+num_std*numpy.std(R),
                               numpy.amax(R)])
        if z_max is None:
            z_max= numpy.amin([numpy.mean(z)+num_std*numpy.std(z),
                               numpy.amax(z)])
        z_min= 0. #Always start grid at z=0 for stars close to plane
        #Separate the coodinates into outliers and normal points
        #Define outliers as points outside of grid
        mask= numpy.any([R < R_min, R > R_max, z > z_max],axis = 0)
        outliers_R= R[mask]
        outliers_z= z[mask]
        normal_R= R[~mask]
        normal_z= z[~mask]
        #Sample the velocity of outliers directly (without interpolation)
        outlier_coord_v= numpy.empty((outliers_R.size, 3))
        for i in range(outliers_R.size):
            outlier_coord_v[i]= self.sampleV(outliers_R[i], outliers_z[i],
                                             use_physical=False)[0]
        #Prepare for optimizing maxVT on a grid
         #Get the new hash of the parameters of grid
        new_hash= hashlib.md5(numpy.array([R_min,R_max,z_max,R_pixel,z_pixel])).hexdigest()
        #Reuse old interpolated object if new hash matches the old one
        if new_hash == self._maxVT_hash:
            ip_max_vT= self._maxVT_ip
        #Generate a new interpolation object if different from before
        else:
            R_number= int((R_max - R_min)/R_pixel)
            z_number= int((z_max - z_min)/z_pixel)
            R_linspace= numpy.linspace(R_min, R_max, R_number)
            z_linspace= numpy.linspace(z_min, z_max, z_number)
            Rv, zv= numpy.meshgrid(R_linspace, z_linspace)
            grid= numpy.dstack((Rv, zv)) #This grid stores (R,z) coordinate
            #Grid is a 3 dimensional array since it stores pairs of values, but 
            #grid max vT is a 2 dimensinal array
            grid_max_vT= numpy.empty((grid.shape[0], grid.shape[1]))
            #Optimize max_vT on the grid
            for i in range(z_number):
                for j in range(R_number):
                    R, z= grid[i][j]
                    grid_max_vT[i][j]= optimize.fmin_powell((lambda x: -self(
                            R,0.,x,z,0.,log=True, use_physical=False)),1.)
            #Determine degree of interpolation
            ky= numpy.min([R_number-1,3])
            kx= numpy.min([z_number-1,3])
            #Generate interpolation object
            ip_max_vT= interpolate.RectBivariateSpline(z_linspace,R_linspace,
                                                       grid_max_vT,kx=kx,ky=ky)
            #Store interpolation object
            self._maxVT_ip= ip_max_vT
            #Update hash of parameters
            self._maxVT_hash= new_hash
        #Evaluate interpolation object to get maxVT at the normal coordinates
        normal_max_vT= ip_max_vT.ev(normal_z, normal_R)
        #Sample all 3 velocities at a normal point and use interpolated vT
        normal_coord_v= \
            self._sampleV_preoptimized(normal_R,normal_z,normal_max_vT)
        #Combine normal and outlier result, preserving original order
        coord_v[mask]= outlier_coord_v
        coord_v[~mask]= normal_coord_v
        if use_physical and not vo is None:
            if _APY_UNITS:
                return units.Quantity(coord_v*vo,unit=units.km/units.s)
            else:
                return coord_v*vo
        else:
            return coord_v

    def _sampleV_preoptimized(self,R,z,maxVT):
        """
        NAME:

           _sampleV_preoptimized

        PURPOSE:

           sample a radial, azimuthal, and vertical velocity at R,z;
           R,z can be an array of positions maxVT is already optimized

        INPUT:

           R - Galactocentric distance (can be Quantity)

           z - height (can be Quantity)
           
           maxVT - an array of pre-optimized maximum vT at corresponding R,z

        OUTPUT:

           a numpy array containing the sampled velocity, (vR, vT, vz),
           where each row correspond to the row of (R,z)

        HISTORY:

           2018-08-09 - Written - Samuel Wong (University of Toronto)

        """
        length = numpy.size(R)
        out= numpy.empty((length,3)) #Initialize output
        #Determine the maximum of the velocity distribution
        maxVR= numpy.zeros(length)
        maxVz= numpy.zeros(length)
        logmaxVD= self(R,maxVR,maxVT,z,maxVz,log=True,use_physical=False)
        #Now rejection-sample
        #Intiialize boolean index of position remaining to be sampled
        remain_indx = numpy.full(length,True)
        while numpy.any(remain_indx):
            nmore= numpy.sum(remain_indx)
            propvR= numpy.random.normal(size=nmore)*2.*self._sr
            propvT= numpy.random.normal(size=nmore)*2.*self._sr+maxVT[remain_indx]
            propvz= numpy.random.normal(size=nmore)*2.*self._sz
            VDatprop= self(R[remain_indx],propvR,propvT,z[remain_indx],propvz,
                           log=True, use_physical=False)-logmaxVD[remain_indx]
            VDatprop-= -0.5*(propvR**2./4./self._sr**2.+
                             propvz**2./4./self._sz**2.+
                             (propvT-maxVT[remain_indx])**2./4./self._sr**2.)
            accept_indx= (VDatprop > numpy.log(numpy.random.random(size=nmore)))
            vR_accept= propvR[accept_indx]
            vT_accept= propvT[accept_indx]
            vz_accept= propvz[accept_indx]
            #Get the indexing of rows of output array that need to be updated
            #with newly accepted velocity
            to_change= numpy.copy(remain_indx)
            to_change[remain_indx]= accept_indx
            out[to_change]= numpy.stack((vR_accept,vT_accept,vz_accept), axis = 1)
            #Removing accepted sampled from remain index
            remain_indx[remain_indx]= ~accept_indx
        return out

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity2',pop=True)
    def pvR(self,vR,R,z,gl=True,ngl=_DEFAULTNGL2,nsigma=4.,vTmax=1.5):
        """
        NAME:

           pvR

        PURPOSE:

           calculate the marginalized vR probability at this location (NOT normalized by the density)

        INPUT:

           vR - radial velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration

           nsigma - sets integration limits to [-1,+1]*nsigma*sigma_z(R) for integration over vz (default: 4)

           vTmax - sets integration limits to [0,vTmax] for integration over vT (default: 1.5)

        OUTPUT:

           p(vR,R,z)

        HISTORY:

           2012-12-22 - Written - Bovy (IAS)

        """
        sigmaz1= self._sz*numpy.exp((self._refr-R)/self._hsz)
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vzgl= nsigma*sigmaz1/2.*(glx+1.)
                vzglw= glw
                vzfac= nsigma*sigmaz1  #2 x integration over [0,nsigma*sigmaz1]
            else:
                vzgl= nsigma*sigmaz1/2.*(glx12+1.)
                vzgl= list(vzgl)
                vzgl.extend(-nsigma*sigmaz1/2.*(glx12+1.))
                vzgl= numpy.array(vzgl)
                vzglw= glw12
                vzglw= list(vzglw)
                vzglw.extend(glw12)
                vzglw= numpy.array(vzglw)
                vzfac = 0.5*nsigma*sigmaz1   #integration over [-nsigma*sigmaz1,0] and [0,nsigma*sigmaz1]
            vTgl= vTmax/2.*(glx+1.)
            vTfac= 0.5 * vTmax  #integration over [0.,vTmax]
            #Tile everything
            vTgl= numpy.tile(vTgl,(ngl,1)).T
            vzgl= numpy.tile(vzgl,(ngl,1))
            vTglw= numpy.tile(glw,(ngl,1)).T #also tile weights
            vzglw= numpy.tile(vzglw,(ngl,1))
            #evaluate
            logqeval= numpy.reshape(self(R+numpy.zeros(ngl*ngl),
                                         vR+numpy.zeros(ngl*ngl),
                                         vTgl.flatten(),
                                         z+numpy.zeros(ngl*ngl),
                                         vzgl.flatten(),
                                         log=True,
                                         use_physical=False),
                                    (ngl,ngl))
            return numpy.sum(numpy.exp(logqeval)*vTglw*vzglw*vzfac)*vTfac

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity2',pop=True)
    def pvT(self,vT,R,z,gl=True,ngl=_DEFAULTNGL2,nsigma=4.):
        """
        NAME:

           pvT

        PURPOSE:

           calculate the marginalized vT probability at this location (NOT normalized by the density)

        INPUT:

           vT - tangential velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration

           nsigma - sets integration limits to [-1,+1]*nsigma*sigma(R) for integration over vz and vR (default: 4)

        OUTPUT:

           p(vT,R,z)

        HISTORY:

           2012-12-22 - Written - Bovy (IAS)
           2018-01-12 - Added Gauss-Legendre integration prefactor nsigma^2/4 - Trick (MPA)

        """
        sigmaR1= self._sr*numpy.exp((self._refr-R)/self._hsr)
        sigmaz1= self._sz*numpy.exp((self._refr-R)/self._hsz)
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vRgl= nsigma*sigmaR1/2.*(glx+1.)
                vzgl= nsigma*sigmaz1/2.*(glx+1.)
                vRglw= glw
                vzglw= glw
                vRfac= nsigma*sigmaR1  #2 x integration over [0,nsigma*sigmaR1]
                vzfac= nsigma*sigmaz1  #2 x integration over [0,nsigma*sigmaz1]
            else:
                vRgl= nsigma*sigmaR1/2.*(glx12+1.)
                vRgl= list(vRgl)
                vRgl.extend(-nsigma*sigmaR1/2.*(glx12+1.))
                vRgl= numpy.array(vRgl)
                vzgl= nsigma*sigmaz1/2.*(glx12+1.)
                vzgl= list(vzgl)
                vzgl.extend(-nsigma*sigmaz1/2.*(glx12+1.))
                vzgl= numpy.array(vzgl)
                vRglw= glw12
                vRglw= list(vRglw)
                vRglw.extend(glw12)
                vRglw= numpy.array(vRglw)
                vzglw= glw12
                vzglw= list(vzglw)
                vzglw.extend(glw12)
                vzglw= numpy.array(vzglw)
                vRfac = 0.5*nsigma*sigmaR1   #integration over [-nsigma*sigmaR1,0] and [0,nsigma*sigmaR1]
                vzfac = 0.5*nsigma*sigmaz1   #integration over [-nsigma*sigmaz1,0] and [0,nsigma*sigmaz1]
            #Tile everything
            vRgl= numpy.tile(vRgl,(ngl,1)).T
            vzgl= numpy.tile(vzgl,(ngl,1))
            vRglw= numpy.tile(vRglw,(ngl,1)).T #also tile weights
            vzglw= numpy.tile(vzglw,(ngl,1))
            #evaluate
            logqeval= numpy.reshape(self(R+numpy.zeros(ngl*ngl),
                                         vRgl.flatten(),
                                         vT+numpy.zeros(ngl*ngl),
                                         z+numpy.zeros(ngl*ngl),
                                         vzgl.flatten(),
                                         log=True,
                                         use_physical=False),
                                    (ngl,ngl))
            return numpy.sum(numpy.exp(logqeval)*vRglw*vzglw*vRfac*vzfac)

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity2',pop=True)
    def pvz(self,vz,R,z,gl=True,ngl=_DEFAULTNGL2,
            nsigma=4.,vTmax=1.5,
            _return_actions=False,_jr=None,_lz=None,_jz=None,
            _return_freqs=False,
            _rg=None,_kappa=None,_nu=None,_Omega=None,
            _sigmaR1=None):
        """
        NAME:

           pvz

        PURPOSE:

           calculate the marginalized vz probability at this location (NOT normalized by the density)

        INPUT:

           vz - vertical velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration

           nsigma - sets integration limits to [-1,+1]*nsigma*sigma_R(R) for integration over vR (default: 4)

           vTmax - sets integration limits to [0,vTmax] for integration over vT (default: 1.5)

        OUTPUT:

           p(vz,R,z)

        HISTORY:

           2012-12-22 - Written - Bovy (IAS)

        """
        if _sigmaR1 is None:
            sigmaR1= self._sr*numpy.exp((self._refr-R)/self._hsr)
        else:
            sigmaR1= _sigmaR1
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vRgl= (glx+1.)
                vRglw= glw
                vRfac= nsigma*sigmaR1  #2 x integration over [0,nsigma*sigmaR1]
            else:
                vRgl= (glx12+1.)
                vRgl= list(vRgl)
                vRgl.extend(-(glx12+1.))
                vRgl= numpy.array(vRgl)
                vRglw= glw12
                vRglw= list(vRglw)
                vRglw.extend(glw12)
                vRglw= numpy.array(vRglw)
                vRfac = 0.5*nsigma*sigmaR1   #integration over [-nsigma*sigmaR1,0] and [0,nsigma*sigmaR1]
            vTgl= vTmax/2.*(glx+1.)
            vTfac= 0.5 * vTmax  #integration over [0.,vTmax]
            #Tile everything
            vTgl= numpy.tile(vTgl,(ngl,1)).T
            vRgl= numpy.tile(vRgl,(ngl,1))
            vTglw= numpy.tile(glw,(ngl,1)).T #also tile weights
            vRglw= numpy.tile(vRglw,(ngl,1))
            #If inputs are arrays, tile
            if isinstance(R,numpy.ndarray):
                nR= len(R)
                R= numpy.tile(R,(ngl,ngl,1)).T.flatten()
                z= numpy.tile(z,(ngl,ngl,1)).T.flatten()
                vz= numpy.tile(vz,(ngl,ngl,1)).T.flatten()
                vTgl= numpy.tile(vTgl,(nR,1,1)).flatten()
                vRgl= numpy.tile(vRgl,(nR,1,1)).flatten()
                vTglw= numpy.tile(vTglw,(nR,1,1))
                vRglw= numpy.tile(vRglw,(nR,1,1))
                scalarOut= False
            else:
                R= R+numpy.zeros(ngl*ngl)
                z= z+numpy.zeros(ngl*ngl)
                vz= vz+numpy.zeros(ngl*ngl)
                nR= 1
                scalarOut= True
                vRgl= vRgl.flatten()
            vRgl*= numpy.tile(nsigma*sigmaR1/2.,(ngl,ngl,1)).T.flatten()
            #evaluate
            if _jr is None and _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self(R,
                                                                 vRgl.flatten(),
                                                                 vTgl.flatten(),
                                                                 z,
                                                                 vz,
                                                                 log=True,
                                                                 _return_actions=True,
                                                                 _return_freqs=True,
                                                                 use_physical=False)
                logqeval= numpy.reshape(logqeval,(nR,ngl*ngl))
            elif not _jr is None and not _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self((_jr,_lz,_jz),
                                           rg=_rg,kappa=_kappa,nu=_nu,
                                           Omega=_Omega,
                                           log=True,
                                           _return_actions=True,
                                           _return_freqs=True,
                                           use_physical=False)
                logqeval= numpy.reshape(logqeval,(nR,ngl*ngl))
            elif not _jr is None and _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self((_jr,_lz,_jz),
                                           log=True,
                                           _return_actions=True,
                                           _return_freqs=True,
                                           use_physical=False)
                logqeval= numpy.reshape(logqeval,(nR,ngl*ngl))
            elif _jr is None and not _rg is None:
                logqeval, jr, lz, jz, rg, kappa, nu, Omega= self(R,
                                             vRgl.flatten(),
                                             vTgl.flatten(),
                                             z,
                                             vz,
                                             rg=_rg,kappa=_kappa,nu=_nu,
                                             Omega=_Omega,
                                             log=True,
                                             _return_actions=True,
                                             _return_freqs=True,
                                             use_physical=False)
                logqeval= numpy.reshape(logqeval,(nR,ngl*ngl))
            vRglw= numpy.reshape(vRglw,(nR,ngl*ngl))
            vTglw= numpy.reshape(vTglw,(nR,ngl*ngl))
            if scalarOut:
                result= numpy.sum(numpy.exp(logqeval)*vTglw*vRglw,axis=1)[0]*vRfac*vTfac
            else:
                result= numpy.sum(numpy.exp(logqeval)*vTglw*vRglw,axis=1)*vRfac*vTfac
            if _return_actions and _return_freqs:
                return (result,
                        jr,lz,jz,
                        rg, kappa, nu, Omega)          
            elif _return_freqs:
                return (result,
                        rg, kappa, nu, Omega)
            elif _return_actions:
                return (result,
                        jr,lz,jz)
            else:
                return result

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity',pop=True)
    def pvRvT(self,vR,vT,R,z,gl=True,ngl=_DEFAULTNGL2,nsigma=4.):
        """
        NAME:

           pvRvT

        PURPOSE:

           calculate the marginalized (vR,vT) probability at this location (NOT normalized by the density)

        INPUT:

           vR - radial velocity (can be Quantity)

           vT - tangential velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration
           
           nsigma - sets integration limits to [-1,+1]*nsigma*sigma_z(R) for integration over vz (default: 4)

        OUTPUT:

           p(vR,vT,R,z)

        HISTORY:

           2013-01-02 - Written - Bovy (IAS)           
           2018-01-12 - Added Gauss-Legendre integration prefactor nsigma/2 - Trick (MPA)

        """
        sigmaz1= self._sz*numpy.exp((self._refr-R)/self._hsz)
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vzgl= nsigma*sigmaz1/2.*(glx+1.)
                vzglw= glw
                vzfac= nsigma*sigmaz1  #2 x integration over [0,nsigma*sigmaz1]
            else:
                vzgl= nsigma*sigmaz1/2.*(glx12+1.)
                vzgl= list(vzgl)
                vzgl.extend(-nsigma*sigmaz1/2.*(glx12+1.))
                vzgl= numpy.array(vzgl)
                vzglw= glw12
                vzglw= list(vzglw)
                vzglw.extend(glw12)
                vzglw= numpy.array(vzglw)
                vzfac = 0.5*nsigma*sigmaz1   #integration over [-nsigma*sigmaz1,0] and [0,nsigma*sigmaz1]
            #evaluate
            logqeval= self(R+numpy.zeros(ngl),
                           vR+numpy.zeros(ngl),
                           vT+numpy.zeros(ngl),
                           z+numpy.zeros(ngl),
                           vzgl,
                           log=True,use_physical=False)
            return numpy.sum(numpy.exp(logqeval)*vzglw*vzfac)

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity',pop=True)
    def pvTvz(self,vT,vz,R,z,gl=True,ngl=_DEFAULTNGL2,nsigma=4.):
        """
        NAME:

           pvTvz

        PURPOSE:

           calculate the marginalized (vT,vz) probability at this location (NOT normalized by the density)

        INPUT:

           vT - tangential velocity (can be Quantity)

           vz - vertical velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration
           
           nsigma - sets integration limits to [-1,+1]*nsigma*sigma_R(R) for integration over vR (default: 4)

        OUTPUT:

           p(vT,vz,R,z)

        HISTORY:

           2012-12-22 - Written - Bovy (IAS)           
           2018-01-12 - Added Gauss-Legendre integration prefactor nsigma/2 - Trick (MPA)

        """
        sigmaR1= self._sr*numpy.exp((self._refr-R)/self._hsr)
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            if isinstance(self._aA,(actionAngle.actionAngleAdiabatic,
                                    actionAngle.actionAngleAdiabaticGrid)):
                vRgl= nsigma*sigmaR1/2.*(glx+1.)
                vRglw= glw
                vRfac= nsigma*sigmaR1  #2 x integration over [0,nsigma*sigmaR1]
            else:
                vRgl= nsigma*sigmaR1/2.*(glx12+1.)
                vRgl= list(vRgl)
                vRgl.extend(-nsigma*sigmaR1/2.*(glx12+1.))
                vRgl= numpy.array(vRgl)
                vRglw= glw12
                vRglw= list(vRglw)
                vRglw.extend(glw12)
                vRglw= numpy.array(vRglw)
                vRfac = 0.5*nsigma*sigmaR1   #integration over [-nsigma*sigmaR1,0] and [0,nsigma*sigmaR1]
            #evaluate
            logqeval= self(R+numpy.zeros(ngl),
                           vRgl,
                           vT+numpy.zeros(ngl),
                           z+numpy.zeros(ngl),
                           vz+numpy.zeros(ngl),
                           log=True,use_physical=False)
            return numpy.sum(numpy.exp(logqeval)*vRglw*vRfac)

    @actionAngle_physical_input
    @physical_conversion('phasespacedensityvelocity',pop=True)
    def pvRvz(self,vR,vz,R,z,gl=True,ngl=_DEFAULTNGL2,vTmax=1.5):
        """
        NAME:

           pvR

        PURPOSE:

           calculate the marginalized (vR,vz) probability at this location (NOT normalized by the density)

        INPUT:

           vR - radial velocity (can be Quantity)

           vz - vertical velocity (can be Quantity)

           R - radius (can be Quantity)

           z - height (can be Quantity)

           gl - use Gauss-Legendre integration (True, currently the only option)

           ngl - order of Gauss-Legendre integration
           
           vTmax - sets integration limits to [0,vTmax] for integration over vT (default: 1.5)

        OUTPUT:

           p(vR,vz,R,z)

        HISTORY:

           2013-01-02 - Written - Bovy (IAS)
           2018-01-12 - Added Gauss-Legendre integration prefactor vTmax/2 - Trick (MPA)

        """
        if gl:
            if ngl % 2 == 1:
                raise ValueError("ngl must be even")
            #Use Gauss-Legendre integration for all
            if ngl == _DEFAULTNGL:
                glx, glw= self._glxdef, self._glwdef
                glx12, glw12= self._glxdef12, self._glwdef12
            elif ngl == _DEFAULTNGL2:
                glx, glw= self._glxdef2, self._glwdef2
                glx12, glw12= self._glxdef, self._glwdef
            else:
                glx, glw= numpy.polynomial.legendre.leggauss(ngl)
                glx12, glw12= numpy.polynomial.legendre.leggauss(ngl//2)
            #Evaluate everywhere
            vTgl= vTmax/2.*(glx+1.)
            vTglw= glw
            vTfac= 0.5 * vTmax  #integration over [0.,vTmax]
            #If inputs are arrays, tile
            if isinstance(R,numpy.ndarray):
                nR= len(R)
                R= numpy.tile(R,(ngl,1)).T.flatten()
                z= numpy.tile(z,(ngl,1)).T.flatten()
                vR= numpy.tile(vR,(ngl,1)).T.flatten()
                vz= numpy.tile(vz,(ngl,1)).T.flatten()
                vTgl= numpy.tile(vTgl,(nR,1)).flatten()
                vTglw= numpy.tile(vTglw,(nR,1))
                scalarOut= False
            else:
                R= R+numpy.zeros(ngl)
                vR= vR+numpy.zeros(ngl)
                z= z+numpy.zeros(ngl)
                vz= vz+numpy.zeros(ngl)
                nR= 1
                scalarOut= True
            #evaluate
            logqeval= numpy.reshape(self(R,
                                         vR,
                                         vTgl,
                                         z,
                                         vz,
                                         log=True,
                                         use_physical=False),
                                    (nR,ngl))
            out= numpy.sum(numpy.exp(logqeval)*vTglw*vTfac,axis=1)
            if scalarOut: return out[0]
            else: return out

    def _calc_epifreq(self,r):
        """
        NAME:
           _calc_epifreq
        PURPOSE:
           calculate the epicycle frequency at r
        INPUT:
           r - radius
        OUTPUT:
           kappa
        HISTORY:
           2012-07-25 - Written - Bovy (IAS@MPIA)
        NOTE:
           takes about 0.1 ms for a Miyamoto-Nagai potential
        """
        return potential.epifreq(self._pot,r)

    def _calc_verticalfreq(self,r):
        """
        NAME:
           _calc_verticalfreq
        PURPOSE:
           calculate the vertical frequency at r
        INPUT:
           r - radius
        OUTPUT:
           nu
        HISTORY:
           2012-07-25 - Written - Bovy (IAS@MPIA)
        NOTE:
           takes about 0.05 ms for a Miyamoto-Nagai potential
        """
        return potential.verticalfreq(self._pot,r)

    def _rg(self,lz):
        """
        NAME:
           _rg
        PURPOSE:
           calculate the radius of a circular orbit of Lz
        INPUT:
           lz - Angular momentum
        OUTPUT:
           radius
        HISTORY:
           2012-07-25 - Written - Bovy (IAS@MPIA)
        NOTE:
           seems to take about ~0.5 ms for a Miyamoto-Nagai potential; 
           ~0.75 ms for a MWPotential
           about the same with or without interpolation of the rotation curve

           Not sure what to do about negative lz...
        """
        if isinstance(lz,numpy.ndarray):
            indx= (lz > self._precomputergLzmax)*(lz < self._precomputergLzmin)
            indxc= True^indx
            out= numpy.empty(lz.shape)
            out[indxc]= self._rgInterp(lz[indxc])
            out[indx]= numpy.array([potential.rl(self._pot,lz[indx][ii]) for ii in range(numpy.sum(indx))])
            return out
        else:
            if lz > self._precomputergLzmax or lz < self._precomputergLzmin:
                return potential.rl(self._pot,lz)
            return numpy.atleast_1d(self._rgInterp(lz))

def _vmomentsurfaceIntegrand(vz,vR,vT,R,z,df,sigmaR1,gamma,sigmaz1,n,m,o): #pragma: no cover because this is too slow; a warning is shown
    """Internal function that is the integrand for the vmomentsurface mass integration"""
    return vR**n*vT**m*vz**o*df(R,vR*sigmaR1,vT*sigmaR1*gamma,z,vz*sigmaz1,
                                use_physical=False)

def _vmomentsurfaceMCIntegrand(vz,vR,vT,R,z,df,sigmaR1,gamma,sigmaz1,mvT,n,m,o):
    """Internal function that is the integrand for the vmomentsurface mass integration"""
    return vR**n*vT**m*vz**o*df(R,vR*sigmaR1,vT*sigmaR1*gamma,z,vz*sigmaz1,use_physical=False)*numpy.exp(vR**2./2.+(vT-mvT)**2./2.+vz**2./2.)

def _jmomentsurfaceIntegrand(vz,vR,vT,R,z,df,sigmaR1,gamma,sigmaz1,n,m,o): #pragma: no cover because this is too slow; a warning is shown
    """Internal function that is the integrand for the vmomentsurface mass integration"""
    return df(R,vR*sigmaR1,vT*sigmaR1*gamma,z,vz*sigmaz1,use_physical=False,
              func= (lambda x,y,z: x**n*y**m*z**o))

def _jmomentsurfaceMCIntegrand(vz,vR,vT,R,z,df,sigmaR1,gamma,sigmaz1,mvT,n,m,o):
    """Internal function that is the integrand for the vmomentsurface mass integration"""
    return df(R,vR*sigmaR1,vT*sigmaR1*gamma,z,vz*sigmaz1,use_physical=False,
              func=(lambda x,y,z: x**n*y**m*z**o))\
              *numpy.exp(vR**2./2.+(vT-mvT)**2./2.+vz**2./2.)



###############################################################################
#      class: actionAngleTorus
#
#             Use McMillan, Binney, and Dehnen's Torus code to calculate (x,v)
#             given actions and angles
#
#
###############################################################################
import warnings
import numpy
from galpy.potential import MWPotential, _isNonAxi
from galpy.util import galpyWarning
from . import actionAngleTorus_c
from .actionAngleTorus_c import _ext_loaded as ext_loaded
from galpy.potential.Potential import _check_c
from galpy.potential.Potential import flatten as flatten_potential
_autofit_errvals= {}
_autofit_errvals[-1]= 'something wrong with input, usually bad starting values for the parameters'
_autofit_errvals[-2]= 'Fit failed the goal by a factor <= 2'
_autofit_errvals[-3]= 'Fit failed the goal by more than 2'
_autofit_errvals[-4]= 'Fit aborted: serious problems occured'
class actionAngleTorus(object):
    """Action-angle formalism using the Torus machinery"""
    def __init__(self,*args,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           initialize an actionAngleTorus object

        INPUT:

           pot= potential or list of potentials (3D)

           tol= default tolerance to use when fitting tori (|dJ|/J)

           dJ= default action difference when computing derivatives (Hessian or Jacobian)

        OUTPUT:

           instance

        HISTORY:

           2015-08-07 - Written - Bovy (UofT)

        """
        if not 'pot' in kwargs: #pragma: no cover
            raise IOError("Must specify pot= for actionAngleTorus")
        self._pot= flatten_potential(kwargs['pot'])
        if _isNonAxi(self._pot):
            raise RuntimeError("actionAngleTorus for non-axisymmetric potentials is not supported")
        if self._pot == MWPotential:
            warnings.warn("Use of MWPotential as a Milky-Way-like potential is deprecated; galpy.potential.MWPotential2014, a potential fit to a large variety of dynamical constraints (see Bovy 2015), is the preferred Milky-Way-like potential in galpy",
                          galpyWarning)
        if ext_loaded:
            self._c= _check_c(self._pot)
            if not self._c:
                raise RuntimeError('The given potential is not fully implemented in C; using the actionAngleTorus code is not supported in pure Python')
        else:# pragma: no cover
            raise RuntimeError('actionAngleTorus instances cannot be used, because the actionAngleTorus_c extension failed to load')
        self._tol= kwargs.get('tol',0.001)
        self._dJ= kwargs.get('dJ',0.001)
        return None
    
    def __call__(self,jr,jphi,jz,angler,anglephi,anglez,**kwargs):
        """
        NAME:

           __call__

        PURPOSE:

           evaluate the phase-space coordinates (x,v) for a number of angles on a single torus

        INPUT:

           jr - radial action (scalar)

           jphi - azimuthal action (scalar)

           jz - vertical action (scalar)

           angler - radial angle (array [N])

           anglephi - azimuthal angle (array [N])

           anglez - vertical angle (array [N])

           tol= (object-wide value) goal for |dJ|/|J| along the torus

        OUTPUT:

           [R,vR,vT,z,vz,phi]

        HISTORY:

           2015-08-07 - Written - Bovy (UofT)

        """
        out= actionAngleTorus_c.actionAngleTorus_xvFreqs_c(\
            self._pot,
            jr,jphi,jz,
            angler,anglephi,anglez,
            tol=kwargs.get('tol',self._tol))
        if out[9] != 0:
            warnings.warn("actionAngleTorus' AutoFit exited with non-zero return status %i: %s" % (out[9],_autofit_errvals[out[9]]),
                          galpyWarning)
        return numpy.array(out[:6]).T

    def xvFreqs(self,jr,jphi,jz,angler,anglephi,anglez,**kwargs):
        """
        NAME:

           xvFreqs

        PURPOSE:

           evaluate the phase-space coordinates (x,v) for a number of angles on a single torus as well as the frequencies

        INPUT:

           jr - radial action (scalar)

           jphi - azimuthal action (scalar)

           jz - vertical action (scalar)

           angler - radial angle (array [N])

           anglephi - azimuthal angle (array [N])

           anglez - vertical angle (array [N])

           tol= (object-wide value) goal for |dJ|/|J| along the torus

        OUTPUT:

           ([R,vR,vT,z,vz,phi],OmegaR,Omegaphi,Omegaz,AutoFit error message)

        HISTORY:

           2015-08-07 - Written - Bovy (UofT)

        """
        out= actionAngleTorus_c.actionAngleTorus_xvFreqs_c(\
            self._pot,
            jr,jphi,jz,
            angler,anglephi,anglez,
            tol=kwargs.get('tol',self._tol))
        if out[9] != 0:
            warnings.warn("actionAngleTorus' AutoFit exited with non-zero return status %i: %s" % (out[9],_autofit_errvals[out[9]]),
                          galpyWarning)
        return (numpy.array(out[:6]).T,out[6],out[7],out[8],out[9])

    def Freqs(self,jr,jphi,jz,**kwargs):
        """
        NAME:

           Freqs

        PURPOSE:

           return the frequencies corresponding to a torus

        INPUT:

           jr - radial action (scalar)

           jphi - azimuthal action (scalar)

           jz - vertical action (scalar)

           tol= (object-wide value) goal for |dJ|/|J| along the torus

        OUTPUT:

           (OmegaR,Omegaphi,Omegaz)

        HISTORY:

           2015-08-07 - Written - Bovy (UofT)

        """
        out= actionAngleTorus_c.actionAngleTorus_Freqs_c(\
            self._pot,
            jr,jphi,jz,
            tol=kwargs.get('tol',self._tol))
        if out[3] != 0:
            warnings.warn("actionAngleTorus' AutoFit exited with non-zero return status %i: %s" % (out[3],_autofit_errvals[out[3]]),
                          galpyWarning)
        return out

    def hessianFreqs(self,jr,jphi,jz,**kwargs):
        """
        NAME:

           hessianFreqs

        PURPOSE:

           return the Hessian d Omega / d J and frequencies Omega corresponding to a torus

        INPUT:

           jr - radial action (scalar)

           jphi - azimuthal action (scalar)

           jz - vertical action (scalar)

           tol= (object-wide value) goal for |dJ|/|J| along the torus

           dJ= (object-wide value) action difference when computing derivatives (Hessian or Jacobian)

           nosym= (False) if True, don't explicitly symmetrize the Hessian (good to check errors)

        OUTPUT:

           (dO/dJ,Omegar,Omegaphi,Omegaz,Autofit error message)

        HISTORY:

           2016-07-15 - Written - Bovy (UofT)

        """
        out= actionAngleTorus_c.actionAngleTorus_hessian_c(\
            self._pot,
            jr,jphi,jz,
            tol=kwargs.get('tol',self._tol),
            dJ=kwargs.get('dJ',self._dJ))
        if out[4] != 0:
            warnings.warn("actionAngleTorus' AutoFit exited with non-zero return status %i: %s" % (out[4],_autofit_errvals[out[4]]),
                          galpyWarning)
        # Re-arrange frequencies and actions to r,phi,z
        out[0][:,:]= out[0][:,[0,2,1]]
        out[0][:,:]= out[0][[0,2,1]]
        if kwargs.get('nosym',False):
            return out
        else :# explicitly symmetrize
            return (0.5*(out[0]+out[0].T),out[1],out[2],out[3],out[4])

    def xvJacobianFreqs(self,jr,jphi,jz,angler,anglephi,anglez,**kwargs):
        """
        NAME:

           xvJacobianFreqs

        PURPOSE:

           return [R,vR,vT,z,vz,phi], the Jacobian d [R,vR,vT,z,vz,phi] / d (J,angle), the Hessian dO/dJ, and frequencies Omega corresponding to a torus at multiple sets of angles

        INPUT:

           jr - radial action (scalar)

           jphi - azimuthal action (scalar)

           jz - vertical action (scalar)

           angler - radial angle (array [N])

           anglephi - azimuthal angle (array [N])

           anglez - vertical angle (array [N])

           tol= (object-wide value) goal for |dJ|/|J| along the torus

           dJ= (object-wide value) action difference when computing derivatives (Hessian or Jacobian)

           nosym= (False) if True, don't explicitly symmetrize the Hessian (good to check errors)

        OUTPUT:

           ([R,vR,vT,z,vz,phi], [N,6] array

            d[R,vR,vT,z,vz,phi]/d[J,angle], --> (N,6,6) array

            dO/dJ, --> (3,3) array

            Omegar,Omegaphi,Omegaz, [N] arrays

            Autofit error message)

        HISTORY:

           2016-07-19 - Written - Bovy (UofT)

        """
        out= actionAngleTorus_c.actionAngleTorus_jacobian_c(\
            self._pot,
            jr,jphi,jz,
            angler,anglephi,anglez,
            tol=kwargs.get('tol',self._tol),
            dJ=kwargs.get('dJ',self._dJ))
        if out[11] != 0:
            warnings.warn("actionAngleTorus' AutoFit exited with non-zero return status %i: %s" % (out[11],_autofit_errvals[out[11]]),
                          galpyWarning)
        # Re-arrange actions,angles to r,phi,z
        out[6][:,:,:]= out[6][:,:,[0,2,1,3,5,4]]
        out[7][:,:]= out[7][:,[0,2,1]]
        out[7][:,:]= out[7][[0,2,1]]
        # Re-arrange x,v to R,vR,vT,z,vz,phi
        out[6][:,:]= out[6][:,[0,3,5,1,4,2]]
        if not kwargs.get('nosym',False):
            # explicitly symmetrize
            out[7][:]= 0.5*(out[7]+out[7].T)
        return (numpy.array(out[:6]).T,out[6],out[7],
                out[8],out[9],out[10],out[11])



###############################################################################
#   surfaceSigmaProfile: classes that implement different surface-mass and 
#                        radial velocity dispersion profiles
# 
#   Includes the following:
#      surfaceSigmaProfile - top-level class that represents a surface 
#                            density profile and a sigma_R profile
#      expSurfaceSigmaProfile - class that represents an exponential surface
#                               density profile and an exponential sigma_R 
#                               profile
###############################################################################
import scipy as sc
class surfaceSigmaProfile(object):
    """Class that contains the surface density and sigma_R^2 profile"""
    def __init__(self):
        """Place holder for implementations of this class"""
        return None

    def formatStringParams(self):
        """
        NAME:
           formatStringParams
        PURPOSE:
           when writing the parameters of this profile, what 
           format-strings to use?
           This function defaults to '%6.4f' for each parameter in self._params
        INPUT:
        OUTPUT:
           tuple of format-strings
        HISTORY:
           2010-03-28 - Written - Bovy (NYU)
        """
        out= []
        for param in self._params:
            out.append('%6.4f')
        return out

    def outputParams(self):
        """
        NAME:
           outputParams
        PURPOSE:'
           return a list of parameters of this profile, to create filenames
        INPUT:
        OUTPUT:
           tuple of parameters (given to % ...)
        HISTORY:
           2010-03-28 - Written - Bovy
        """
        return tuple(self._params)

    def surfacemass(self,R,log=False):
        """
        NAME:
           surfacemass
        PURPOSE:
           return the surface density profile at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the log (default: False)
        OUTPUT:
           Sigma(R)
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        raise NotImplementedError("'surfacemass' function not implemented for this surfaceSigmaProfile class")

    def sigma2(self,R,log=False):
        """
        NAME:
           sigma2
        PURPOSE:
           return the radial velocity variance at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the log (default: False)
        OUTPUT:
           sigma^2(R)
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        raise NotImplementedError("'sigma2' function not implemented for this surfaceSigmaProfile class")
        
class expSurfaceSigmaProfile(surfaceSigmaProfile):
    """Exponential surface density and sigma_R^2 class"""
    def __init__(self,params=(1./3.,1.0,0.2)):
        """
        NAME:
           __init__
        PURPOSE:
           initialize an exponential surface density and sigma_R^2 profile
        INPUT:
           params - tuple/list of (surface scale-length,sigma scale-length,
                    sigma(ro)) (NOTE: *not* sigma2 scale-length)
        OUTPUT:
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        surfaceSigmaProfile.__init__(self)
        self._params= params

    def surfacemass(self,R,log=False):
        """
        NAME:
           surfacemass
        PURPOSE:
           return the surface density profile at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the log (default: False)
        OUTPUT:
           Sigma(R)
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        if log:
            return -R/self._params[0]
        else:
            return sc.exp(-R/self._params[0])

    def surfacemassDerivative(self,R,log=False):
        """
        NAME:
           surfacemassDerivative
        PURPOSE:
           return the derivative wrt R of the surface density profile at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the derivative of the log (default: False)
        OUTPUT:
           Sigma'(R) or (log Sigma(r) )'
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        if log:
            return -1./self._params[0]
        else:
            return -sc.exp(-R/self._params[0])/self._params[0]

    def sigma2(self,R,log=False):
        """
        NAME:
           sigma2
        PURPOSE:
           return the radial velocity variance at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the log (default: False)
        OUTPUT:
           sigma^2(R)
        HISTORY:
           2010-03-26 - Written - Bovy (NYU)
        """
        if log:
            return 2.*sc.log(self._params[2])-2.*(R-1.)/self._params[1]
        else:
            return self._params[2]**2.*sc.exp(-2.*(R-1.)/self._params[1])

    def sigma2Derivative(self,R,log=False):
        """
        NAME:
           sigmaDerivative
        PURPOSE:
           return the derivative wrt R of the sigma_R^2 profile at this R
        INPUT:
           R - Galactocentric radius (/ro)
           log - if True, return the derivative of the log (default: False)
        OUTPUT:
           Sigma_R^2'(R) or (log Sigma_R^2(r) )'
        HISTORY:
           2011-03-24 - Written - Bovy (NYU)
        """
        if log:
            return -2./self._params[1]
        else:
            return self._params[2]**2.*sc.exp(-2.*(R-1.)/self._params[1])\
                *(-2./self._params[1])



import sys
import distutils.sysconfig as sysconfig
import warnings
import numpy as nu
from scipy import integrate
import ctypes
import ctypes.util
from numpy.ctypeslib import ndpointer
import os
from galpy import potential
from galpy.util import galpyWarning
from ..util.multi import parallel_map
from .integratePlanarOrbit import _parse_integrator, _parse_tol
from .integrateFullOrbit import _parse_pot as _parse_pot_full
from ..potential.linearPotential import _evaluatelinearForces
from galpy.potential.verticalPotential import verticalPotential
from galpy.potential.WrapperPotential import parentWrapperPotential
from ..util.leung_dop853 import dop853
from ..util import bovy_symplecticode as symplecticode
#Find and load the library
_lib= None
outerr= None
PY3= sys.version > '3'
if PY3:
    _ext_suffix= sysconfig.get_config_var('EXT_SUFFIX')
else: #pragma: no cover
    _ext_suffix= '.so'
for path in sys.path:
    try:
        _lib = ctypes.CDLL(os.path.join(path,'galpy_integrate_c%s' % _ext_suffix))
    except OSError as e:
        if os.path.exists(os.path.join(path,'galpy_integrate_c%s' % _ext_suffix)): #pragma: no cover
            outerr= e
        _lib = None
    else:
        break
if _lib is None: #pragma: no cover
    if not outerr is None:
        warnings.warn("integrateLinearOrbit_c extension module not loaded, because of error '%s' " % outerr,
                      galpyWarning)
    else:
        warnings.warn("integrateLinearOrbit_c extension module not loaded, because galpy_integrate_c%s image was not found" % _ext_suffix,
                      galpyWarning)
    _ext_loaded= False
else:
    _ext_loaded= True

def _parse_pot(pot):
    """Parse the potential so it can be fed to C"""
    from .integrateFullOrbit import _parse_scf_pot
    #Figure out what's in pot
    if not isinstance(pot,list):
        pot= [pot]
    #Initialize everything
    pot_type= []
    pot_args= []
    npot= len(pot)
    for p in pot:
        # Prepare for wrappers NOT CURRENTLY SUPPORTED, SEE PLANAR OR FULL
        if isinstance(p,verticalPotential) \
                and isinstance(p._Pot,potential.MN3ExponentialDiskPotential):
            # Need to do this one separately, because combination of many parts
            # Three Miyamoto-Nagai disks
            npot+= 2
            pot_type.extend([5,5,5])
            pot_args.extend([p._Pot._amp*p._Pot._mn3[0]._amp,
                             p._Pot._mn3[0]._a,p._Pot._mn3[0]._b,
                             p._R,p._phi,
                             p._Pot._amp*p._Pot._mn3[1]._amp,
                             p._Pot._mn3[1]._a,p._Pot._mn3[1]._b,
                             p._R,p._phi,
                             p._Pot._amp*p._Pot._mn3[2]._amp,
                             p._Pot._mn3[2]._a,p._Pot._mn3[2]._b,
                             p._R,p._phi])
        elif isinstance(p,verticalPotential) \
                and isinstance(p._Pot,potential.DiskSCFPotential):
            # Need to do this one separately, because combination of many parts
            # Need to pull this apart into: (a) SCF part, (b) constituent
            # [Sigma_i,h_i] parts
            # (a) SCF, multiply in any add'l amp
            pt,pa= _parse_scf_pot(p._Pot._scf,extra_amp=p._Pot._amp)
            pot_type.append(pt)
            pot_args.extend(pa)
            pot_args.extend([p._R,p._phi])
            # (b) constituent [Sigma_i,h_i] parts
            for Sigma,hz in zip(p._Pot._Sigma_dict,p._Pot._hz_dict):
                npot+= 1
                pot_type.append(26)
                stype= Sigma.get('type','exp')
                if stype == 'exp' \
                        or (stype == 'exp' and 'Rhole' in Sigma):
                    pot_args.extend([3,0,
                                     4.*nu.pi*Sigma.get('amp',1.)*p._Pot._amp,
                                     Sigma.get('h',1./3.)])
                elif stype == 'expwhole' \
                        or (stype == 'exp' and 'Rhole' in Sigma):
                    pot_args.extend([4,1,
                                     4.*nu.pi*Sigma.get('amp',1.)*p._Pot._amp,
                                     Sigma.get('h',1./3.),
                                     Sigma.get('Rhole',0.5)])
                hztype= hz.get('type','exp')
                if hztype == 'exp':
                    pot_args.extend([0,hz.get('h',0.0375)])
                elif hztype == 'sech2':
                    pot_args.extend([1,hz.get('h',0.0375)])
                pot_args.extend([p._R,p._phi])
        elif isinstance(p,potential.KGPotential):
            pot_type.append(31)
            pot_args.extend([p._amp,p._K,p._D2,2.*p._F]) 
        elif isinstance(p,potential.IsothermalDiskPotential):
            pot_type.append(32)
            pot_args.extend([p._amp*p._sigma2/p._H,2.*p._H]) 
        # All other potentials can be handled in the same way as follows:
        elif isinstance(p,verticalPotential):
            _,pt,pa= _parse_pot_full(p._Pot)
            pot_type.extend(pt)
            pot_args.extend(pa)
            pot_args.append(p._R)
            pot_args.append(p._phi)
    pot_type= nu.array(pot_type,dtype=nu.int32,order='C')
    pot_args= nu.array(pot_args,dtype=nu.float64,order='C')
    return (npot,pot_type,pot_args)

def integrateLinearOrbit_c(pot,yo,t,int_method,rtol=None,atol=None,dt=None):
    """
    NAME:
       integrateLinearOrbit_c
    PURPOSE:
       C integrate an ode for a LinearOrbit
    INPUT:
       pot - Potential or list of such instances
       yo - initial condition [q,p], can be [N,2] or [2]
       t - set of times at which one wants the result
       int_method= 'leapfrog_c', 'rk4_c', 'rk6_c', 'symplec4_c'
       rtol, atol
       dt= (None) force integrator to use this stepsize (default is to automatically determine one; only for C-based integrators)
    OUTPUT:
       (y,err)
       y : array, shape (N,len(t),2) or (len(y0),len(t)) if N=1
       Array containing the value of y for each desired time in t, \
       with the initial value y0 in the first row.
       err: error message, if not zero: 1 means maximum step reduction happened for adaptive integrators
    HISTORY:
       2018-10-06 - Written - Bovy (UofT)
       2018-10-14 - Adapted to allow multiple orbits to be integrated at once - Bovy (UofT)
    """
    if len(yo.shape) == 1: single_obj= True
    else: single_obj= False
    yo= nu.atleast_2d(yo)
    nobj= len(yo)
    rtol, atol= _parse_tol(rtol,atol)
    npot, pot_type, pot_args= _parse_pot(pot)
    int_method_c= _parse_integrator(int_method)
    if dt is None: 
        dt= -9999.99

    #Set up result array
    result= nu.empty((nobj,len(t),2))
    err= nu.zeros(nobj,dtype=nu.int32)

    #Set up the C code
    ndarrayFlags= ('C_CONTIGUOUS','WRITEABLE')
    integrationFunc= _lib.integrateLinearOrbit
    integrationFunc.argtypes= [ctypes.c_int,
                               ndpointer(dtype=nu.float64,flags=ndarrayFlags),
                               ctypes.c_int,                             
                               ndpointer(dtype=nu.float64,flags=ndarrayFlags),
                               ctypes.c_int,
                               ndpointer(dtype=nu.int32,flags=ndarrayFlags),
                               ndpointer(dtype=nu.float64,flags=ndarrayFlags),
                               ctypes.c_double,
                               ctypes.c_double,
                               ctypes.c_double,
                               ndpointer(dtype=nu.float64,flags=ndarrayFlags),
                               ndpointer(dtype=nu.int32,flags=ndarrayFlags),
                               ctypes.c_int]

    #Array requirements, first store old order
    f_cont= [yo.flags['F_CONTIGUOUS'],
             t.flags['F_CONTIGUOUS']]
    yo= nu.require(yo,dtype=nu.float64,requirements=['C','W'])
    t= nu.require(t,dtype=nu.float64,requirements=['C','W'])
    result= nu.require(result,dtype=nu.float64,requirements=['C','W'])
    err= nu.require(err,dtype=nu.int32,requirements=['C','W'])

    #Run the C code
    integrationFunc(ctypes.c_int(nobj),
                    yo,
                    ctypes.c_int(len(t)),
                    t,
                    ctypes.c_int(npot),
                    pot_type,
                    pot_args,
                    ctypes.c_double(dt),
                    ctypes.c_double(rtol),ctypes.c_double(atol),
                    result,
                    err,
                    ctypes.c_int(int_method_c))
    
    if nu.any(err == -10): #pragma: no cover
        raise KeyboardInterrupt("Orbit integration interrupted by CTRL-C (SIGINT)")

    #Reset input arrays
    if f_cont[0]: yo= nu.asfortranarray(yo)
    if f_cont[1]: t= nu.asfortranarray(t)

    if single_obj: return (result[0],err[0])
    else: return (result,err)

# Python integration functions
def integrateLinearOrbit(pot,yo,t,int_method,rtol=None,atol=None,numcores=1,
                         dt=None):
    """
    NAME:
       integrateLinearOrbit
    PURPOSE:
       Integrate an ode for a LinearOrbit
    INPUT:
       pot - Potential or list of such instances
       yo - initial condition [q,p], shape [N,2]
       t - set of times at which one wants the result
       int_method= 'leapfrog', 'odeint', or 'dop853'
       rtol, atol= tolerances (not always used...)
       numcores= (1) number of cores to use for multi-processing
       dt= (None) force integrator to use this stepsize (default is to automatically determine one; only for C-based integrators)
    OUTPUT:
       (y,err)
       y : array, shape (N,len(t),2)
       Array containing the value of y for each desired time in t, \
       with the initial value y0 in the first row.
       err: error message, always zero for now
    HISTORY:
       2010-07-13- Written - Bovy (NYU)
       2019-04-08 - Adapted to allow multiple orbits to be integrated at once and moved to integrateLinearOrbit.py - Bovy (UofT)
    """
    if int_method.lower() == 'leapfrog':
        if rtol is None: rtol= 1e-8
        def integrate_for_map(vxvv):
            return symplecticode.leapfrog(lambda x,t=t: \
                                              _evaluatelinearForces(pot,x,t=t),
                                          nu.array(vxvv),
                                          t,rtol=rtol)
    elif int_method.lower() == 'dop853':
        if rtol is None: rtol= 1e-8
        def integrate_for_map(vxvv):
            return dop853(func=_linearEOM,x=vxvv,t=t,args=(pot,))
    elif int_method.lower() == 'odeint':
        if rtol is None: rtol= 1e-8
        def integrate_for_map(vxvv):
            return integrate.odeint(_linearEOM,vxvv,t,args=(pot,),rtol=rtol)
    else: # Assume we are forcing parallel_mapping of a C integrator...
        def integrate_for_map(vxvv):
            return integrateLinearOrbit_c(pot,nu.copy(vxvv),
                                          t,int_method,dt=dt)[0]
    if len(yo) == 1: # Can't map a single value...
        return nu.atleast_3d(integrate_for_map(yo[0]).T).T, 0
    else:
        return (nu.array((parallel_map(integrate_for_map,yo,numcores=numcores))),
                nu.zeros(len(yo)))

def _linearEOM(y,t,pot):
    """
    NAME:
       linearEOM
    PURPOSE:
       the one-dimensional equation-of-motion
    INPUT:
       y - current phase-space position
       t - current time
       pot - (list of) linearPotential instance(s)
    OUTPUT:
       dy/dt
    HISTORY:
       2010-07-13 - Bovy (NYU)
    """
    return [y[1],_evaluatelinearForces(pot,y[0],t=t)]



from __future__ import division, print_function

import os, os.path
import copy
import pickle
import numpy as nu
import galpy.util.bovy_plot as plot
from galpy.util import config
from .Potential import PotentialError, flatten
from galpy.util.bovy_conversion import physical_conversion,\
    potential_physical_input
_APY_LOADED= True
try:
    from astropy import units
except ImportError:
    _APY_LOADED= False
class linearPotential(object):
    """Class representing 1D potentials"""
    def __init__(self,amp=1.,ro=None,vo=None):
        self._amp= amp
        self.dim= 1
        self.isRZ= False
        self.hasC= False
        # Parse ro and vo
        if ro is None:
            self._ro= config.__config__.getfloat('normalization','ro')
            self._roSet= False
        else:
            if _APY_LOADED and isinstance(ro,units.Quantity):
                ro= ro.to(units.kpc).value
            self._ro= ro
            self._roSet= True
        if vo is None:
            self._vo= config.__config__.getfloat('normalization','vo')
            self._voSet= False
        else:
            if _APY_LOADED and isinstance(vo,units.Quantity):
                vo= vo.to(units.km/units.s).value
            self._vo= vo
            self._voSet= True
        return None

    def __mul__(self,b):
        """
        NAME:

           __mul__

        PURPOSE:

           Multiply a linearPotential's amplitude by a number

        INPUT:

           b - number

        OUTPUT:

           New instance with amplitude = (old amplitude) x b

        HISTORY:

           2019-01-27 - Written - Bovy (UofT)

        """
        if not isinstance(b,(int,float)):
            raise TypeError("Can only multiply a planarPotential instance with a number")
        out= copy.deepcopy(self)
        out._amp*= b
        return out
    # Similar functions
    __rmul__= __mul__
    def __div__(self,b): return self.__mul__(1./b)
    __truediv__= __div__

    def __add__(self,b):
        """
        NAME:

           __add__

        PURPOSE:

           Add linearPotential instances together to create a multi-component potential (e.g., pot= pot1+pot2+pot3)

        INPUT:

           b - linearPotential instance or a list thereof

        OUTPUT:

           List of linearPotential instances that represents the combined potential

        HISTORY:

           2019-01-27 - Written - Bovy (UofT)

        """
        if isinstance(b,list):
            return [self]+b
        else:
            return [self,b]
    # Define separately to keep order
    def __radd__(self,b):
        if isinstance(b,list):
            return b+[self]
        else:
            raise TypeError("Can only add a Force or Potential instance to another instance or to a list of such instances")

    def turn_physical_off(self):
        """
        NAME:

           turn_physical_off

        PURPOSE:

           turn off automatic returning of outputs in physical units

        INPUT:

           (none)

        OUTPUT:

           (none)

        HISTORY:

           2016-01-30 - Written - Bovy (UofT)

        """
        self._roSet= False
        self._voSet= False
        return None

    def turn_physical_on(self,ro=None,vo=None):
        """
        NAME:

           turn_physical_on

        PURPOSE:

           turn on automatic returning of outputs in physical units

        INPUT:

           ro= reference distance (kpc; can be Quantity)

           vo= reference velocity (km/s; can be Quantity)

        OUTPUT:

           (none)

        HISTORY:

           2016-01-30 - Written - Bovy (UofT)

        """
        self._roSet= True
        self._voSet= True
        if not ro is None:
            if _APY_LOADED and isinstance(ro,units.Quantity):
                ro= ro.to(units.kpc).value
            self._ro= ro
        if not vo is None:
            if _APY_LOADED and isinstance(vo,units.Quantity):
                vo= vo.to(units.km/units.s).value
            self._vo= vo
        return None

    @potential_physical_input
    @physical_conversion('energy',pop=True)
    def __call__(self,x,t=0.):
        """
        NAME:
           __call__
        PURPOSE:

           evaluate the potential

        INPUT:

           x - position (can be Quantity)

           t= time (optional; can be Quantity)

        OUTPUT:

           Phi(x,t)

        HISTORY:

           2010-07-12 - Written - Bovy (NYU)

        """
        return self._call_nodecorator(x,t=t)

    def _call_nodecorator(self,x,t=0.):
        # Separate, so it can be used during orbit integration
        try:
            return self._amp*self._evaluate(x,t=t)
        except AttributeError: #pragma: no cover
            raise PotentialError("'_evaluate' function not implemented for this potential")

    @potential_physical_input
    @physical_conversion('force',pop=True)
    def force(self,x,t=0.):
        """
        NAME:

           force

        PURPOSE:

           evaluate the force

        INPUT:

           x - position (can be Quantity)

           t= time (optional; can be Quantity)

        OUTPUT:

           F(x,t)

        HISTORY:

           2010-07-12 - Written - Bovy (NYU)

        """
        return self._force_nodecorator(x,t=t)

    def _force_nodecorator(self,x,t=0.):
        # Separate, so it can be used during orbit integration
        try:
            return self._amp*self._force(x,t=t)
        except AttributeError: #pragma: no cover
            raise PotentialError("'_force' function not implemented for this potential")

    def plot(self,t=0.,min=-15.,max=15,ns=21,savefilename=None):
        """
        NAME:

           plot

        PURPOSE:

           plot the potential

        INPUT:

           t - time to evaluate the potential at

           min - minimum x

           max - maximum x

           ns - grid in x

           savefilename - save to or restore from this savefile (pickle)

        OUTPUT:

           plot to output device

        HISTORY:

           2010-07-13 - Written - Bovy (NYU)

        """
        if not savefilename == None and os.path.exists(savefilename):
            print("Restoring savefile "+savefilename+" ...")
            savefile= open(savefilename,'rb')
            potx= pickle.load(savefile)
            xs= pickle.load(savefile)
            savefile.close()
        else:
            xs= nu.linspace(min,max,ns)
            potx= nu.zeros(ns)
            for ii in range(ns):
                potx[ii]= self._evaluate(xs[ii],t=t)
            if not savefilename == None:
                print("Writing savefile "+savefilename+" ...")
                savefile= open(savefilename,'wb')
                pickle.dump(potx,savefile)
                pickle.dump(xs,savefile)
                savefile.close()
        return plot.bovy_plot(xs,potx,
                              xlabel=r"$x/x_0$",ylabel=r"$\Phi(x)$",
                              xrange=[min,max])

@potential_physical_input
@physical_conversion('energy',pop=True)
def evaluatelinearPotentials(Pot,x,t=0.):
    """
    NAME:

       evaluatelinearPotentials

    PURPOSE:

       evaluate the sum of a list of potentials

    INPUT:

       Pot - (list of) linearPotential instance(s)

       x - evaluate potentials at this position (can be Quantity)

       t - time to evaluate at  (can be Quantity)

    OUTPUT:

       pot(x,t)

    HISTORY:

       2010-07-13 - Written - Bovy (NYU)

    """
    return _evaluatelinearPotentials(Pot,x,t=t)

def _evaluatelinearPotentials(Pot,x,t=0.):
    """Raw, undecorated function for internal use"""
    if isinstance(Pot,list):
        sum= 0.
        for pot in Pot:
            sum+= pot._call_nodecorator(x,t=t)
        return sum
    elif isinstance(Pot,linearPotential):
        return Pot._call_nodecorator(x,t=t)
    else: #pragma: no cover
        raise PotentialError("Input to 'evaluatelinearPotentials' is neither a linearPotential-instance or a list of such instances")

@potential_physical_input
@physical_conversion('force',pop=True)
def evaluatelinearForces(Pot,x,t=0.):
    """
    NAME:

       evaluatelinearForces

    PURPOSE:

       evaluate the forces due to a list of potentials

    INPUT:

       Pot - (list of) linearPotential instance(s)

       x - evaluate forces at this position (can be Quantity)

       t - time to evaluate at (can be Quantity)

    OUTPUT:

       force(x,t)

    HISTORY:

       2010-07-13 - Written - Bovy (NYU)

    """
    return _evaluatelinearForces(Pot,x,t=t)

def _evaluatelinearForces(Pot,x,t=0.):
    """Raw, undecorated function for internal use"""
    if isinstance(Pot,list):
        sum= 0.
        for pot in Pot:
            sum+= pot._force_nodecorator(x,t=t)
        return sum
    elif isinstance(Pot,linearPotential):
        return Pot._force_nodecorator(x,t=t)
    else: #pragma: no cover
        raise PotentialError("Input to 'evaluateForces' is neither a linearPotential-instance or a list of such instances")

def plotlinearPotentials(Pot,t=0.,min=-15.,max=15,ns=21,savefilename=None):
    """
    NAME:

       plotlinearPotentials

    PURPOSE:

       plot a combination of potentials

    INPUT:

       t - time to evaluate potential at

       min - minimum x

       max - maximum x

       ns - grid in x

       savefilename - save to or restore from this savefile (pickle)

    OUTPUT:

       plot to output device

    HISTORY:

       2010-07-13 - Written - Bovy (NYU)

    """
    Pot= flatten(Pot)
    if not savefilename == None and os.path.exists(savefilename):
        print("Restoring savefile "+savefilename+" ...")
        savefile= open(savefilename,'rb')
        potx= pickle.load(savefile)
        xs= pickle.load(savefile)
        savefile.close()
    else:
        xs= nu.linspace(min,max,ns)
        potx= nu.zeros(ns)
        for ii in range(ns):
            potx[ii]= evaluatelinearPotentials(Pot,xs[ii],t=t)
        if not savefilename == None:
            print("Writing savefile "+savefilename+" ...")
            savefile= open(savefilename,'wb')
            pickle.dump(potx,savefile)
            pickle.dump(xs,savefile)
            savefile.close()
    return plot.bovy_plot(xs,potx,
                          xlabel=r"$x/x_0$",ylabel=r"$\Phi(x)$",
                          xrange=[min,max])




###############################################################################
#   MovingObjectPotential.py: class that implements the potential coming from
#                             a moving object
###############################################################################
import copy
import numpy as nu
from .Potential import Potential, _isNonAxi, flatten, \
    evaluatePotentials, evaluateRforces, evaluatezforces, evaluateDensities, _check_c
from .PlummerPotential import PlummerPotential
class MovingObjectPotential(Potential):
    """
    Class that implements the potential coming from a moving object by combining
    any galpy potential with an integrated galpy orbit.
    """
    def __init__(self,orbit,pot=None,amp=1.0,
                 ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a MovingObjectPotential

        INPUT:

           orbit - the Orbit of the object (Orbit object)

           pot - A potential object or list of potential objects representing the potential of the moving object; should be spherical, but this is not checked [default= PlummerPotential(amp=0.06,b=0.01)]
           
           amp (=1.) another amplitude to apply to the potential

           ro=, vo= distance and velocity scales for translation into internal units (default from configuration file)

        OUTPUT:

           (none)

        HISTORY:

           2011-04-10 - Started - Bovy (NYU)

           2018-10-18 - Re-implemented to represent general object potentials using galpy potential models - James Lane (UofT)

        """

        Potential.__init__(self,amp=amp,ro=ro,vo=vo)       
        # If no potential supplied use a default Plummer sphere
        if pot is None:
            pot=PlummerPotential(amp=0.06,b=0.01)
            self._pot = pot
        else:
            pot=flatten(pot)
            if _isNonAxi(pot):
                raise NotImplementedError('MovingObjectPotential for non-axisymmetric potentials is not currently supported')
            self._pot=pot
        self._orb= copy.deepcopy(orbit)
        self._orb.turn_physical_off()
        self.isNonAxi= True
        self.hasC= _check_c(self._pot)
        return None

    def _evaluate(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,z, phi
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,z,phi)
        HISTORY:
           2011-04-10 - Started - Bovy (NYU)
           2018-10-18 - Updated for general object potential - James Lane (UofT)
        """
        #Cylindrical distance
        Rdist = _cylR(R,phi,self._orb.R(t),self._orb.phi(t))
        orbz = self._orb.z(t) if self._orb.dim() == 3 else 0
        #Evaluate potential
        return evaluatePotentials( self._pot, Rdist, orbz-z, use_physical=False)

    def _Rforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        HISTORY:
           2011-04-10 - Written - Bovy (NYU)
           2018-10-18 - Updated for general object potential - James Lane (UofT)
        """
        #Cylindrical distance
        Rdist = _cylR(R,phi,self._orb.R(t),self._orb.phi(t))
        # Difference vector
        orbz = self._orb.z(t) if self._orb.dim() == 3 else 0
        (xd,yd,zd) = _cyldiff(self._orb.R(t), self._orb.phi(t), orbz,
            R, phi, z)
        #Evaluate cylindrical radial force
        RF = evaluateRforces(self._pot,Rdist,zd, use_physical=False)

        # Return R force, negative of radial vector to evaluation location.
        return -RF*(nu.cos(phi)*xd+nu.sin(phi)*yd)/Rdist

    def _zforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _zforce
        PURPOSE:
           evaluate the vertical force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the vertical force
        HISTORY:
           2011-04-10 - Written - Bovy (NYU)
           2018-10-18 - Updated for general object potential - James Lane (UofT)
        """
        #Cylindrical distance
        Rdist = _cylR(R,phi,self._orb.R(t),self._orb.phi(t))
        # Difference vector
        orbz = self._orb.z(t) if self._orb.dim() == 3 else 0
        (xd,yd,zd) = _cyldiff(self._orb.R(t), self._orb.phi(t), orbz,
            R, phi, z)
        #Evaluate and return z force
        return -evaluatezforces(self._pot,Rdist,zd, use_physical=False)

    def _phiforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _phiforce
        PURPOSE:
           evaluate the azimuthal force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the azimuthal force
        HISTORY:
           2011-04-10 - Written - Bovy (NYU)
           2018-10-18 - Updated for general object potential - James Lane (UofT)
        """
        #Cylindrical distance
        Rdist = _cylR(R,phi,self._orb.R(t),self._orb.phi(t))
        # Difference vector
        orbz = self._orb.z(t) if self._orb.dim() == 3 else 0
        (xd,yd,zd) = _cyldiff(self._orb.R(t), self._orb.phi(t), orbz,
            R, phi, z)
        #Evaluate cylindrical radial force.
        RF = evaluateRforces(self._pot, Rdist, zd, use_physical=False)
        # Return phi force, negative of phi vector to evaluate location
        return -RF*R*(nu.cos(phi)*yd-nu.sin(phi)*xd)/Rdist

    def _dens(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _dens
        PURPOSE:
           evaluate the density for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the density
        HISTORY:
           2010-08-08 - Written - Bovy (NYU)
        """
        #Cylindrical distance
        Rdist = _cylR(R,phi,self._orb.R(t),self._orb.phi(t))
        # Difference vector
        orbz = self._orb.z(t) if self._orb.dim() == 3 else 0
        (xd,yd,zd) = _cyldiff(self._orb.R(t), self._orb.phi(t), orbz,
            R, phi, z)
        # Return the density
        return evaluateDensities(self._pot, Rdist, zd, use_physical=False)

def _cylR(R1,phi1,R2,phi2):
    return nu.sqrt(R1**2.+R2**2.-2.*R1*R2*nu.cos(phi1-phi2)) # Cosine law

def _cyldiff(R1,phi1,z1,R2,phi2,z2):
    dx = R1*nu.cos(phi1)-R2*nu.cos(phi2)
    dy = R1*nu.sin(phi1)-R2*nu.sin(phi2)
    dz = z1-z2
    return (dx,dy,dz)
